/* mapcgi.c
 *
 * Buster SSI routines for mapping network
 *
 * ROUTINES: br_portimage(), br_vcimage(), br_minpkts(), 
 * ROUTINES: br_hourpkts(), br_ipimage(), maphost(), ht_netrtr(), 
 * ROUTINES: br_sorthosts(), ht_netmap(), br_formhost(), br_hostparm(), 
 * ROUTINES: ht_hostimg(), ht_showhost(), br_vcnlines(), next_cn(), 
 * ROUTINES: name_gt(), br_sortcns(), mapcon(), map_cnhost(), hostrow(), 
 * ROUTINES: ht_connmap(), ht_recentlist
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */

#include "busterport.h"   /* Praemio */
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brdns.h"

#include "websys.h"
#include "webio.h"
#include "webfs.h"
#include "wsfdata.h"


extern   char *   br_getname(struct ip_hostentry * iphost);
extern   char *   act_color(int pkts, int interval, u_long metric);


/* the congroup structure contains info for one or more connections to a given
 * port & remote IP address.
 */
#define MAPCONNS  128      /* number of connection groups for a single host */
#define MAXCONNS  8        /* number of connections per group */
#define MAXVCONNS 10       /* number of connections per vertical set */

struct congroup
{
   struct conn_entry * conn[MAXCONNS];  /* First N matching connections */
   int      count;      /* number of matching connections */
   unshort  port;
   int      m_flags;    /* bitmask of M_ bits, below */
   ip_addr  ipaddr;
};

#define M_ISLOCAL 0x01  /* "remote" IP is local (on LAN) */
#define M_INET    0x04  /* Internet (not local or ISP) */


/* get color string for idle item (white->grey) */

char * 
idlecolors[] = 
{
   "recent1",
   "recent2",
   "recent3",
   "recent4",
   "awhile1",
   "awhile2",
   "awhile3",
   "awhile4",
   "aging1",
   "aging2",
   "aging3",
   "aging4",
   "idle1",
   "idle2",
   "idle3",
   "idle4",
};

/* idle_color()
 * 
 * Return a class name to control for idle time background color
 * 
 */

/* Table with time (seconds) to color mapping */
#define S_HOUR 3600
#define S_DAY  (S_HOUR * 24)
static u_long sec_table[16] = 
{ 
   5, 10, 20, 40,
   70, 120, 240, 600, 
   1200, S_HOUR, S_HOUR * 2, S_HOUR * 4, 
   S_HOUR * 8, S_DAY, S_DAY * 2, S_DAY * 7
};  

char *
idle_color(u_long idle) /* ctricks unit has been idle */
{
   int i;

   idle /= TPS;   /* Convert idle time to seconds */

   /* Find time in seconds table, use to index color table. */
   for(i = 0; i < 16; i++)
   {
      if(idle < sec_table[i])
         return ( idlecolors[i] );
   }

   return ( idlecolors[15] );
}

char * portimages[] =
{
   "cnred.gif",         /* 0 - red */
   "cnorange.gif",      /* 1 - orange */
   "cnyellow.gif",      /* 2 - yellow */
   "cngreen.gif",       /* 3 - green */
   "cnltblue.gif",      /* 4 - light blue */
   "cnblue.gif",        /* 5 - blue */
   "cnpurple.gif",      /* 6 - purple */
   "cnpink.gif",        /* 7 - magenta */
};

char vcfilebuf[14];  /* tmp buffer for the names above */
#define NUMPCOLORS (sizeof(portimages)/sizeof(char*))

/* next two routines return names of connection image file based 
 * on port values 
 */

char *
br_portimage(unshort port)
{
   switch(port)
   {
   case 80:    /* http */
      return portimages[4];   /* Light Blue */
   case 135:   /* Netbios */
   case 137:
   case 138:
   case 139:
   case 445:
   case 1026:
      return portimages[0];   /* Red */
   case 21:    /* ftp */
   case 19:    /* telnet */
      return portimages[6];   /* Purple */
   case 110:   /* POP */
      return portimages[7];   /* Magenta */
   case 53:    /* DNS */
      return portimages[5];   /* Blue */
   case 25:    /* SMTP */
      return portimages[2];   /* Yellow */
   default:
      if(port <= 1024)
         return portimages[3];   /* Green */
      else
         return(portimages[1]);  /* Orange */
   }
}

char *
br_vcimage(unshort port)
{
   char * cp;

   cp = br_portimage(port);
   if(!cp)
      panic("vcimage");

   /* change the "cn" file name to a "vc" file name */
   strcpy(vcfilebuf, cp);
   vcfilebuf[0] = 'v';
   vcfilebuf[1] = 'c';
   return &vcfilebuf[0];
}

/* br_idleinfo() 
 *
 * Set gui control info for colored "idle time" cell
 *
 * Returns style name, fills in timeunits and timecount
 */

char *
br_idleinfo(u_long idle, char ** timeunits, int * timecount )
{

   /* figure out what time units to use and set variables */
   if(idle < (2 * MINUTE))    /* Less than 2 minutes -> seconds */
   {
      *timeunits = "seconds";
      *timecount = idle/SECOND;
   }
   else if(idle < (2 * HOUR))  /* Less than 2 hours -> minutes */
   {
      *timeunits = "minutes";
      *timecount = idle/MINUTE;
   }
   else if(idle < (2 * DAY))  /* Less than 2 days -> hours */
   {
      *timeunits = "hours";
      *timecount = idle/HOUR;
   }
   else
   {
      *timeunits = "days";
      *timecount = idle/DAY;
   }

   return(idle_color(idle));
}


/* HTML strings for map of local net host */

char * netmap3 = "<img border=0 src=\"%s\" height=50 width=70></a>";

char * netmap4 = "</td><td class=\"%s\" width=40 height=11 align=right "
   "title=\"%lu Packets in last hour\" >%lu</td></tr>";

char * netmap5 = "<tr><td></td><td class=\"%s\"  height=11 align=right "
   "onMouseover=\"dropdownmenu(this, event, tt_connct, '180px')\" "
   "onMouseout=\"delayhidemenu()\" >\n"
   "%lu </td></tr>\n";

char * idlerow = "<tr><td></td><td class=\"%s\" width=40 height=11 align=right "
   "title=\"%u %s since last network activity\" >%u</td></tr>\n";

char * idlerow0 = "<tr><td></td><td class=\"%s\" width=40 height=11 align=right "
   "title=\"Unknown time since last network activity\" >%s</td></tr>\n";

char * favicon = "<img src=\"http://%u.%u.%u.%u/favicon.ico\" border=0 height=22 width=22> ";



/* br_minpkts() - returns number of packets host has seen in the past minute */

u_long
br_minpkts(struct host_entry * host)
{
   int i;
   u_long   ct = 0;

   for(i = 0; i < 60; i++)
      ct += host->sec_p[i];

   return ct;
}

/* br_hourpkts() - returns number of packets host has seen in the past hour */

u_long
br_hourpkts(struct host_entry * host)
{
   int i;
   u_long   ct = 0;

   for(i = 0; i < 60; i++)       /* count pkts from last 60 minutes */
      ct += host->min_p[i];
   return ct;
}

char *
br_ipimage(struct ip_hostentry * iphost)
{
   if(iphost->imagefile[0])
      return(iphost->imagefile);
   if(iphost->flags & IH_BROADCAST)
      return "bcast.gif";
   if(iphost->flags & IH_MULTICAST)
      return "bcast.gif";
   if(ipe_lookup(htonl(iphost->ipaddr)))  /* Is IP address on blacklist? */
      return "ipblock.gif";
   if(iphost->flags & IH_SERVER)
       return "iphost.gif";
   return "unknown.gif";
}

/* maphost() - output HTML to display a host in "map" format.
 *
 * Only one of the last three parameters needs to be passed. 
 * - If host is passed then it's values are used; providing the most detail
 * - If iphost is passed then it's somewhat less detailed info is used.
 * - If neither is passed, then we assume this may be a target IP which 
 * has not responded and an "all unknown" picture is displayed, using the 
 * IP address passed (which may conceivable be zero).
 */

void
maphost(wi_sess * sess, 
        struct host_entry * host,
        struct ip_hostentry * iphost,
        struct congroup * cn)
{
   u_long      pkts;    /* packet sent in past minute */
   u_long      idle;    /* number of ticks host has been idle */
   ip_addr     ipaddr;
   char *      cp;
   char *      color;
   char *      timestyle;
   char *      tooltip;
   char *      timeunits;  /* Minutes, seconds, hours, etc. */
   int         timecount;  /* Idle time in "timeunits" */
   int         namelen;
   int         inethost = FALSE;
   int         local = FALSE;
   int         conns;      /* Scratch */

   if(host)
   {
      local = TRUE;

      /* If host has never sent a packet it was probably read from 
       * a network.db file. We have to guess the idle time.
       */
      if(host->active == 0)
      {
         /* If br_sleeptime is zero, this was a coldstart */
         if(br_sleeptime)
         {
            idle = (br_sleeptime/TPS) * HOUR;
         }
         else  /* coldstart */
         {
            /* This really should be "don't know", but since we don't 
             * have a color for that we fudge to a few hours */
            idle = cticks/TPS;       /* start with uptime */
            if(idle < (8 * HOUR))
               idle = 8 * HOUR;  /* Make it at lease few hours */
         }
      }
      else
      {
         idle = cticks - host->active;    /* how long host has been idle */
      }
      if(iphost)  /* caller passed both */
      {
         dtrap();
      }
      iphost = host->iphost;
      if(iphost)
         ipaddr = iphost->ipaddr;
      else
         ipaddr = 0;
   }
   else if(iphost)
   {
      inethost = TRUE;

      /* see if the IP host has an associated MAC host */
      host = iphost->mac_host;
      ipaddr = iphost->ipaddr;
      idle = cticks - iphost->tm_lastpkt;    /* how long host has been idle */
   }
   else
   {
      ipaddr = cn->conn[0]->rem.ce_addr;
      idle = cticks - cn->conn[0]->ce_time;
   }

   timestyle = br_idleinfo(idle, &timeunits, &timecount );

   wi_printf(sess, "<table border=0 bgcolor=#FFFFFF align=center "
      "cellpadding=0 cellspacing=1 ><tr align=center >");

   /* first take the case where we have a local host that is NOT routed */
   if(host && (host->iphost == iphost))
   {
      /* See if we need a "you are here" or "Buster click" icon */
      if(host->h_flags & H_SELF)
      {
         wi_printf(sess, "<td title=\"You are here\" >"
            "<img src=\"youhere.gif\" width=16 height=16 ></td>");
      }
      else if(iphost && iphost->buster)
      {
         wi_printf(sess, "<td title=\"Buster on this machine, click to link\" >"
            "<a href=\"http://%u.%u.%u.%u:%u\"  target='top' >"
            "<img src=\"favicon.ico\" border='0' width=16 height=16 ></a></td>",
            PASS_IPADDR(iphost->ipaddr), iphost->buster->httpport );
      }
      else     /* No icon, just make empty cell */
         wi_printf(sess, "<td></td>");

      wi_printf(sess, "<td rowspan=3 title=\"Click to see connections for this Computer\" >\n");
      wi_printf(sess, "<a href=\"connmap.htm?host=%s\">\n",
         print_mac(host->macaddr) );

      /* select image for this machine */
      wi_printf(sess, netmap3, br_image(host) );

      /* print colored indexed traffic cell. */
      pkts = br_hourpkts(host);
      wi_printf(sess, netmap4, act_color(1, HOUR, pkts), pkts, pkts);

      /* print colored indexed connections cell */
      if(iphost)
         conns = iphost->conns;
      else
         conns = 0;

      wi_printf(sess, netmap5, act_color(1, MINUTE, iphost->conns), conns, conns);

      /* Print idle time cell */
      if(pkts == 0)
         wi_printf(sess, idlerow0, timestyle, "?" );
      else
         wi_printf(sess, idlerow, timestyle, timecount, timeunits, timecount );
   }
   else if(iphost)  /* print image and stats based on IP host */
   {
      u_long seconds;      /* seconds host has been monitored */


      /* See if we can provide a "rating" icon */
      if(iphost->ratingHtml[0])
      {
         wi_printf(sess, 
            "<td title=\"Click to see SiteAdvisor page for this site\" >%s</td>", 
            iphost->ratingHtml);
      }
      else
      {
         wi_printf(sess, "<td></td>");
      }

      wi_printf(sess, "<td rowspan=3 >"
         "<span onMouseover=\"dropdownmenu(this, event, tt_whois, '180px')\" "
         "onMouseout=\"delayhidemenu()\" >\n");

      /* Print the link for clicking on the host */
      wi_printf(sess, "<a href=\"hostinfo.htm?host=%u.%u.%u.%u\">\n",
         PASS_IPADDR(ipaddr) );

      /* select image for this machine */
      wi_printf(sess, netmap3, br_ipimage(iphost) );

      /* End span with whois tooltip */
      wi_printf(sess, "</span>\n");

      seconds = ((cticks - iphost->tm_detected) /TPS);
      if(seconds > HOUR)
         seconds = HOUR;   /* limit statistics to past hour */
      else if (seconds < 1)
         seconds = 1;      /* prevent divide by zero */

      pkts = iphost->tot_pkts;
      wi_printf(sess, netmap4, act_color(1, HOUR, pkts), pkts, pkts);
      conns = iphost->conns;
      wi_printf(sess, netmap5, act_color(1, HOUR, conns), conns, conns);
      wi_printf(sess, idlerow, timestyle, timecount, timeunits, timecount );
   }
   else
   {
      wi_printf(sess, "<td></td>");   /* upper cell to left of image */

      wi_printf(sess, "<td rowspan=3 title=\"Click to see connections for this Computer\" >\n");
      wi_printf(sess, "<a href=\"hostinfo.htm?host=%u.%u.%u.%u\">\n",
         PASS_IPADDR(ipaddr) );
      wi_printf(sess, netmap3, "unknown.gif" );

      pkts = cn->count;
      wi_printf(sess, netmap4, act_color(1, HOUR, pkts), pkts, pkts);
      wi_printf(sess, netmap5, act_color(1, HOUR, pkts), pkts, pkts);
      wi_printf(sess, idlerow, timestyle, timecount, timeunits, timecount );
   }

   if(host && local)
      cp = br_getmacname(host);
   else
      cp = br_getname(iphost);

   if(local && host && (host->h_flags & H_DUPLIP))
   {
      struct host_entry * tmp_host;
      int   i;

      /* Check that this is still a duplicate */
      for(i = 0; i < HASH_HOSTS; i++)
      {
         tmp_host = hosttab[i];
         while(tmp_host)
         {
            if((tmp_host->iphost) &&
               (tmp_host->iphost->ipaddr == ipaddr) &&
               (tmp_host != host))
            {
               goto dup_ip;
            }
            tmp_host = tmp_host->next;
         }
      }
      host->h_flags &= ~H_DUPLIP;
dup_ip:
      if(host->h_flags & H_DUPLIP)   /* still a duplicate IP? */
      {
         color = "FF3322";
         tooltip = "tt_dupip";
      }
      else
      {
         host->h_flags &= ~H_DUPLIP;
         color = "FFFFFF";
         tooltip = "tt_ipaddr";
      }
   }
   else     /* IP address is not a duplicate */
   {
      color = "FFFFFF";
      tooltip = "tt_ipaddr";
   }

   if(host && local)
   {
      wi_printf(sess, "<tr><td colspan=3 align=center >"
         "<span  onMouseover=\"dropdownmenu(this, event, tt_locname, '180px')\" "
         "onMouseout=\"delayhidemenu()\" >");
   }
   else
   {
      wi_printf(sess, "<tr><td colspan=3 align=center >"
         "<span  onMouseover=\"dropdownmenu(this, event, tt_domname, '180px')\" "
         "onMouseout=\"delayhidemenu()\" >");
   }

   /* make a direct link to server around this name */
   if(!local)
   {
      ip_addr ipaddr;
      char *  linkname;

      /* if it's dot notation, make a URL */
      if((br_checkip(&ipaddr, cp)) == NULL)
         wi_printf(sess, "<a href=\"http://%s\" target=top>", cp );
      else     /* text name, clip leading subdomains */
      {
         linkname = strrchr(cp, '.');  /* find last dot */
         if(!linkname)  
         {
            /* no dots in name, use IP address */
            if(iphost)
            {
               wi_printf(sess, "<a href=\"http://%u.%u.%u.%u\" target=top>", 
                  PASS_IPADDR(iphost->ipaddr) );
            }
         }
         else
         {
            /* back up to second last dot, that's our name */
            while((linkname > cp) && ((*(linkname - 1)) != '.'))
               linkname--;
            wi_printf(sess, "<a href=\"http://%s\" target=top>", linkname );
         }
      }
   }

   namelen = strlen(cp);
   /* Reduce font size for overly long names */
   if(namelen < 22)
   {
      wi_printf(sess, "%s", cp );
   }
   else
   {
      if(namelen >= (MAXDNSNAME-1))    /* extremely long = bug? */
      {
         dtrap();
         *(cp + (MAXDNSNAME-1)) = 0;     /* truncate it */
         namelen = (MAXDNSNAME-1);
      }
      wi_printf(sess, "<font size=-%d>%s</font>", (namelen>30)?2:1, cp  );
   }
   if(!local)
   {
      wi_printf(sess, "</a>" );
   }

   wi_printf(sess, 
      "</span></td></tr>\n<tr><td colspan=3 align=center bgcolor=#%s >"
      "<span onMouseover=\"dropdownmenu(this, event, %s, '180px')\" "
      "onMouseout=\"delayhidemenu()\" >",
      color, tooltip );

   if(iphost && iphost->flags & IH_FAVICON)
      wi_printf(sess, favicon, PASS_IPADDR(ipaddr) );

   wi_printf(sess, "<font size=-1 >%u.%u.%u.%u</span></td></tr>", 
      PASS_IPADDR(ipaddr) );

   if(local)      /* if local host, display mac address */
   {
      /* use pkts and idle vars to hold 24 bit MAC chuncks */
      if(host)
      {
         cp = print_mac(&host->macaddr[0]);
      }
      else
      {
         cp = "(unknown MAC)";
      }

      wi_printf(sess, "<tr><td colspan=3 align=center >"
         "<span onMouseover=\"dropdownmenu(this, event, tt_macaddr, '180px')\" "
         "onMouseout=\"delayhidemenu()\" >\n");

      /* If we have mac info, put mac in machost link */
      if(host)
      {
         wi_printf(sess, "<a href=\"machost.htm?host=%s\" >",
            print_mac(&host->macaddr[0]) );
      }

      wi_printf(sess, "<font size=-2>%s</font>%s</span>",
         cp, host?"</a>":"" );

      wi_printf(sess, "</td></tr></table>\n");
   }
   else
      wi_printf(sess, "</table>\n");


   return;
}


/* ht_netrtr()
 *
 * SSI routine.
 *
 * Show the network router only.
 *
 * Returns 0 if OK. 
 */

int
ht_netrtr(wi_sess * sess, EOFILE * eofile)
{
   int      othernet;
   struct host_entry * router;

   router = brstates[br_ifview].br_router;

   wi_printf(sess, "<table border=0 width=100%c >\n<tr><td width=23%c >", '%', '%');
   wi_printf(sess, "<font color=#888888 size=-1 >Watching network <b>\"%s\"</b></font>",
      nets[br_ifview]->description);

   if(nets[br_ifview]->ifOperStatus != 1)
      wi_printf(sess, "<font color=#FF2244 ><b> - DISABLED </b></font>");
   
   wi_printf(sess, "</td><td width=10%c></td><td width=33%c><font face=arial color=#888888 >Router:</font>",
      '%', '%');
   if(router)
      maphost(sess, router, NULL, 0);
   else
      wi_printf(sess, "No router found yet");
   wi_printf(sess, "</td><td width=10%c></td><td width=33%c aligh=left >", '%', '%');
   if(ifNumber > 2)
   {
      wi_printf(sess, "<a href='setnet.htm?net=-1'>switch to another network adapter</a>");
   }
   else if(ifNumber == 2)
   {
      /* Offer the other interface by description */
      if(br_ifview == 0)
         othernet = 1;
      else
         othernet = 0;
      wi_printf(sess, "<font color=#888888  size=-1 >\n");
      wi_printf(sess, "Switch to network adapter <a href='setnet.htm?net=%d'>\n", othernet);
      wi_printf(sess, "%s</a>\n", nets[othernet]->description);
   }
   wi_printf(sess, "</td></tr></table>\n");

   return 0; /* OK return */
}

/* br_sorthosts() - helper routine for ht_netmap()
 * 
 * This returns an array of all the MAC hosts, sorted as per the global 
 * variable br_cfg.sortparm. Number of hosts is returned in count.
 *
 * The returned pointer MUST be freed by the caller, via npfree();
 *
 * Returns NULL if out of memory.
 */

struct host_entry **
br_sorthosts(int * count)
{
   int      i;             /* index */
   int      hosts = 0;     /* number of hosts */
   struct host_entry *  host;          /* scratch host */
   struct host_entry ** sort;          /* scratch host */
   struct host_entry ** host_array;    /* sorted hosts */
   struct host_entry ** end;           /* for insertion sort */

   /* Allocate host_array based on the recorded host count plus 2.*/
   host_array = (struct host_entry **)
      npalloc((brstates[br_ifview].machosts + 2) * sizeof(void*));
   if(host_array == NULL)
   {
      dtrap();
      return NULL;
   }
   /* To count and sort the hosts, loop through hosts table, doing 
    * insertion sort into the alloced array. 
    */
   for(i = 0; i < HASH_HOSTS; i++)
   {
      for(host = hosttab[i]; host; host = host->next)
      {
         if(host->iface != br_ifview)     /* only deal with the viewable iface */
            continue;

         if(host == brstates[br_ifview].br_router)  /* Is this host the router? */
         {
            /* skip router, it's already done. */
            continue;
         }
         if(host->h_flags & H_BCAST)   /* skip the broadcast entry */
            continue;
         if(host->macaddr[0] & 0x01)   /* check the Multicast bit */
            continue;

         if((host->iphost == NULL) || (host->iphost->mac_host != host))
         {
//            dtrap();    /* Corrupt entry? */
            continue;
         }

         /* see if we should filter out this host - later */

         for(sort = host_array; *sort; sort++)
         {
            if(br_cfg.sortparm == 0)     /* no sort parameter */
            {
               /* just put host at the end of the sort list */
               sort = host_array + hosts;
               break;
            }
            if(br_cfg.sortparm == 'i')
            {
               /* Make sure we didn't timeout an IP host. This is most 
                * likely to happen coming back from hiberbnation. 
                */
               if(((*sort)->iphost == NULL) ||
                  (host->iphost == NULL))
               {
                  dtrap();
                  npfree(host_array);
                  return NULL;
               }
               if((*sort)->iphost->ipaddr > host->iphost->ipaddr)
               {
                  /* found place to insert host; move rest of list back 1 slot */
                  end = (host_array + hosts);
                  while(end > sort)
                  {
                     *end = *(end - 1);
                     end--;
                  }
                  break;   /* out of for(sort = host_array... */
               }
            }
            else if(br_cfg.sortparm == 't')
            {
               /* see if host's count is greater than sort's count */
               if( br_hourpkts(host) > br_hourpkts(*sort))
               {
                  /* found place to insert host; move rest of list back 1 slot */
                  end = (host_array + hosts);
                  while(end > sort)
                  {
                     *end = *(end - 1);
                     end--;
                  }
                  break;   /* out of for(sort = host_array... */
               }
            }
            /* check for bogus sort parm */
            if((br_cfg.sortparm != 'i') && 
               (br_cfg.sortparm != 't') && 
               (br_cfg.sortparm != 'I') && 
               (br_cfg.sortparm != 's') && 
               (br_cfg.sortparm != 'n') && 
               (br_cfg.sortparm != 'p'))
            {
               dtrap();
            }
         }
         /* "sort" now points to slot for host in sorted list */
         *sort = host;
         hosts++;
      }
   }

   /* host array now filled with sorted hosts; */
   *count = hosts;
   return host_array;
}


/* ht_netmap()
 *
 * SSI routine.

   This draws the master map of the whole network.

   Returns 0 if OK. 
*/

int
ht_netmap(wi_sess * sess, EOFILE * eofile)
{
   int      i;             /* index */
   int      col_left = 0;  /* map table columns, init to avoid compiler warnings */
   int      hosts = 0;     /* number of hosts */
   struct host_entry *  host;          /* scratch host */
   struct host_entry ** host_array;    /* sorted hosts */

   if(brstates[br_ifview].machosts == 0)
   {
      wi_printf(sess, "No hosts discovered yet");
      return 0;
   }

   /* get array containing all the MAC hosts */
   host_array = br_sorthosts(&hosts);
   if(!host_array)
   {
      wi_printf(sess, "Housecleaning, please try again.\n");
      return 0;
   }

   /* Generate table header. 
    * - Table is hardcoded for 4 columns to simplify the line graphic layout logic. 
    * - We insert a slim 5th col (cell) between cols 2 & 3 in case we need to draw a 
    *   vertical network connecting line.
    * - The stupid trick with the '%c' is to get around a bug in the Microsoft C 6.0 
    *   printf engine. 
    * - If we print row/col counts in the table tag, MS IE 6.0 does the WRONG THING 
    *   with col widths, so don't add code to do this.
    */
   wi_printf(sess, "<table width=100%c border=0 cellspacing=0 cellpadding=0>\n", '%');

   /* Print a thin spacing row - 4 24% width cells for the computers, with a 
    * thin 4% cell in the middle for the vertical gray bar
    */
   wi_printf(sess, "<tr><td height=10 width=24%c></td><td width=24%c></td>"
      "<td width=4%c class=\"netmapvertical\" ></td>"
      "<td width=24%c></td><td width=24%c></td></tr> \n", '%', '%', '%', '%', '%');

   for(i = 0; i < hosts; i++)
   {
      /* prior to every row of 4 hosts, print a grey network connection line. This has to be 
       * horizontally sized based on the number of hosts in this row, 1-4. We do this by 
       * making a three-cell table, with the gray bar in the middle cell and adjusting the
       * three cell widths to get the bar where we want it.
       */
      if((i % 4) == 0)
      {
         char * tab_start = "<tr><td colspan=5 ><table border=0 cellspacing=0 cellpadding=0 "
            "width=100%c ><tr><td width=%u%c></td>\n";
         char * tab_mid = "<td width=%u%c  title=\"The gray bar represents your network\" "
            " class=\"netmapacross\"><img src=\"left-end.gif\" class=\"lefty\">"
            "<img src=\"right-end.gif\" class=\"righty\">\n";
         char * tab_end = "<td width=%u%c ></td></tr></table></td></tr>\n";


         col_left = hosts - i;        /* number of populated columns remaining */

         if(col_left >= 4)    /* stretch across all 4 hosts */
         {
            wi_printf(sess, tab_start, '%', '%', 11, '%' );
            wi_printf(sess, tab_mid, 78, '%' );
            wi_printf(sess, tab_end, 11, '%' );
         }
         else if(col_left == 3)  /* stretch across first 3 hosts */
         {
            wi_printf(sess, tab_start, '%', '%', 11, '%' );
            wi_printf(sess, tab_mid, 54, '%' );
            wi_printf(sess, tab_end, 35, '%' );
         }
         else if(col_left == 2)  /* stretch across middle 2 hosts */
         {
            wi_printf(sess, tab_start, '%', '%', 35, '%' );
            wi_printf(sess, tab_mid, 30, '%' );
            wi_printf(sess, tab_end, 35, '%' );
         }
         else if(col_left == 1)
         {
            wi_printf(sess, tab_start, '%', '%', 35, '%' );
            wi_printf(sess, tab_mid, 16, '%' );
            wi_printf(sess, tab_end, 49, '%' );
         }
         else
         {
            dtrap();
            return -1;
         }

         /* Start next row of hosts */
         wi_printf(sess, "\n<tr><td>");

         /* if there's only 1 or 2 items in this row, generate an empty placeholder cell */
         if(col_left <= 2)
         {
            wi_printf(sess, "&nbsp;</td><td>");  /* start next cell */
         }
      }

      /* print individual hosts */
      host = *(host_array + i);      /* get next host from sorted list */

      /* build a table around the host with riser grey bar */
      wi_printf(sess, "<table border=0 align=center cellspacing=0 cellpadding=0 >"
         "<tr><td class=\"netmapvertical\">&nbsp;</td></tr><tr><td>\n");

      maphost(sess, host, NULL, 0);
      wi_printf(sess, "</td></tr><tr height=15 ><td></td></tr></table>\n");

      /* put outer table cell boundary. */
      if((i % 4) == 3)     /* hosts 3,7,11,etc, get new row */
      {
         wi_printf(sess, "</td></tr>\n");
      }
      else     /* not end-of-row cell */
      {
         int col = i & 0x03;     /* cols are numbered 0, 1, 2, 3 */

         /* In middle, insert a thin cell for vertical network line */
         if(((col_left > 2) && (col == 1)) ||   /* after 2nd col */
            ((col_left <= 2) && (col == 0)))    /* last line, after 1st col */
         {
            if(col_left > 4)
               wi_printf(sess, "</td><td class=\"netmapvertical\">&nbsp;");
            else
               wi_printf(sess, "</td><td><br>");  /* invisable content helps formatting */
         }
         wi_printf(sess, "</td>\n<td>");  /* start next cell */
      }
   }
   /* If last computer is on a line by itself, add a cell to even up table */
   if(i%4 == 1)
      wi_printf(sess, "</td><td><br>");

   wi_printf(sess, "</td></tr></table>\n");

   npfree(host_array);

   return 0; /* OK return */
}


/* parse_mac() - parse a text MAC address into the passed byuffer.
 *  
 *  buffer must be at least 6 bytes 
 *  
 *  returns 0 if OK, -1 if format error.
 */

int
parse_mac(char * mac,   /* text for MAC address */
          u_char * out)   /* bninary buffer for output */
{
   int   i;
   char * cp = mac;

   for(i = 0; i < 6; i++)
   {
      if(((*(cp+2) != '-')) && (i > 5))
         return -1;
      *out = (u_char)atoh(cp);

      if((*out == 0) && *cp != '0')
         return -1;

      out++;
      cp += 3;
   }
   return 0;
}


/* br_formhost() - get the MAC and IP hosts structures for the IP address 
 * specified in a form. name/value format is "host=%u.%u.%u.%u"for IP
 * hosts or xx-xx-xx-xx-xx-xx for MAC hosts.
 *
 * This returns 0 if both were set, -1 if one or neither were set.
 */

int
br_formhost(wi_sess * sess,          /* IN, has "host" entry */
            struct host_entry ** host,       /* OUT */
            struct ip_hostentry ** iphost)   /* OUT */
{
   char *   cp;
   ip_addr  ipaddr;
   struct host_entry * machost;

   /* get the mac & IP host structures, with error checking */
   cp = wi_formvalue(sess, "host");
   if(!cp)
      { dtrap(); return -1; }

   if((*(cp+2) == '-') ||   /* MAC host format */
      (*(cp+2) == ' '))
   {
      u_char macaddr[6];
      parse_mac(cp, &macaddr[0]);
      machost = br_hostlookup((char*)macaddr, br_ifview);
      if(!machost)
         return -1;
      *host = machost;
      *iphost = machost->iphost;
      return 0;
   }
   ipaddr = inet_addr(cp);
   *iphost = br_ipanyif(ipaddr);
   if(!(*iphost))
      { dtrap(); return -1; }
   *host = (*iphost)->mac_host;
   if(!(*host))
   {
      /* this is OK if all the caller wanted was the IP host */
      return -1; 
   }

   return 0;      /* OK return */
}


int
br_hostparm(wi_sess * sess, char * args)
{
   int      i;          /* scratch index */
   u_long   code;       /* first 4 chars of arg */
   u_long   ct;         /* scratch counter */
   struct ip_hostentry * iphost;
   struct host_entry * machost;

   /* Make "code" from first 4 chars of next arg */
   code = *args;
   code = (code << 8) + *(args + 1);
   code = (code << 8) + *(args + 2);
   code = (code << 8) + *(args + 3);

   /* Get the mac and IP host */
   i = br_formhost(sess, &machost, &iphost);
   if(i != 0)
      return -1;

   switch (code)
   {
   case 0x6e616d65:  /* name */
      wi_printf(sess, br_getmacname(machost) );
      break;
   case 0x6d616361:  /* macaddr */
      wi_printf(sess, "%s", print_mac(machost->macaddr) );
      break;
   case 0x69705F31:  /* ip_1 */
      wi_printf(sess, "%u.%u.%u.%u", PASS_IPADDR(machost->iphost->ipaddr) );
      break;
   case 0x69705F32:  /* ip_2 */
      wi_printf(sess, "%u.%u.%u.%u", PASS_IPADDR(machost->ipaddr2) );
      break;
   case 0x6c656173:  /* lease */
      if(machost->ip_lease == -1)
         wi_printf(sess, "Infinite");
      else
         wi_printf(sess, "%lu sec.", machost->ip_lease );
      break;
   case 0x726f7574:  /* router */
      wi_printf(sess, "%s", (machost->h_flags & H_ROUTER)?"Yes":"No" );
      break;

   case 0x735f6370:  /* s_cpkts */
      wi_printf(sess, "class=\"%s\" ", act_color(1, SECOND, machost->sec_pkts) );
      break;
   case 0x735f706b:  /* s_pkts */
      wi_printf(sess, "%lu", machost->sec_pkts );
      break;
   case 0x735f6362:  /* s_cbytes - color for bytes */
      wi_printf(sess, "class=\"%s\" ", act_color(0, SECOND, machost->sec_bytes) );
      break;
   case 0x735f6279:  /* s_bytes */
      wi_printf(sess, "%lu", machost->sec_bytes );
      break;
   case 0x6d5f6370:  /* m_cpkts */
      ct = br_minpkts(machost);
      wi_printf(sess, "class=\"%s\" ", act_color(1, MINUTE, ct) );
      break;
   case 0x6d5f706b:  /* m_pkts */
      ct = br_minpkts(machost);
      wi_printf(sess, "%lu", ct );
      break;
   case 0x6d5f6362:  /* m_cbytes */
      ct = 0;
      for(i = 0; i < 60; i++)
         ct += machost->sec_b[i];
      wi_printf(sess, "class=\"%s\" ", act_color(0, MINUTE, ct) );
      break;
   case 0x6d5f6279:  /* m_bytes */
      ct = 0;
      for(i = 0; i < 60; i++)
         ct += machost->sec_b[i];
      wi_printf(sess, "%lu", ct );
      break;
   case 0x685f6370:  /* h_cpkts */
      ct = br_hourpkts(machost);
      wi_printf(sess, "class=\"%s\" ", act_color(1, HOUR, ct) );
      break;
   case 0x685f706b:  /* h_pkts */
      ct = br_hourpkts(machost);
      wi_printf(sess, "%lu", ct );
      break;
   case 0x685f6362:  /* h_cbytes */
      ct = 0;
      for(i = 0; i < 60; i++)
         ct += machost->min_b[i];
      wi_printf(sess, "class=\"%s\" ", act_color(0, HOUR, ct) );
      break;
   case 0x685f6279:  /* h_bytes */
      ct = 0;
      for(i = 0; i < 60; i++)
         ct += machost->min_b[i];
      wi_printf(sess, "%lu", ct );
      break;
   case 0x645f6370:  /* d_cpkts */
      ct = 0;
      for(i = 0; i < 24; i++)
         ct += machost->hour_p[i];
      wi_printf(sess, "class=\"%s\" ", act_color(1, DAY, ct) );
      break;
   case 0x645f706b:  /* d_pkts */
      ct = 0;
      for(i = 0; i < 24; i++)
         ct += machost->hour_p[i];
      wi_printf(sess, "%lu", ct);
      break;
   case 0x645f6362:  /* d_cbytes */
      ct = 0;
      for(i = 0; i < 24; i++)
         ct += machost->hour_b[i];
      wi_printf(sess, "class=\"%s\" ", act_color(0, DAY, ct) );
      break;
   case 0x645f6279:  /* d_bytes */
      ct = 0;
      for(i = 0; i < 24; i++)
         ct += machost->hour_b[i];
      wi_printf(sess, "%lu", ct );
      break;
   default:
      dtrap();    /* bad html */
      return -1;
   }
   return 0;
}


/* ht_hostimg()
 *
 * SSI routine.
 *
 * Get the file name for the image for passed IP address
 *
 * Returns 0 if OK. 
 */

int
ht_hostimg(wi_sess * sess, EOFILE * eofile)
{
   int      err;
   struct ip_hostentry * iphost;
   struct host_entry * machost;

   /* Get the mac and IP host */
   err = br_formhost(sess, &machost, &iphost);
   if(err != 0)
   {
      /* See if error means we have IP host info only */
      if((machost == NULL) && (iphost))
      {
         if(iphost->imagefile[0])
            wi_printf(sess, "%s", iphost->imagefile);   
         else
            wi_printf(sess, "unknown.gif");
         return  0;
      }
      return -1;
   }

   wi_printf(sess, br_image(machost));   

   return 0; /* OK return */
}




/* ht_showhost() - SSI routine.
 *
 * display a single local (MAC) host, with minimal stats
 *
 * Returns 0 if OK. 
 */

int
ht_showhost(wi_sess * sess, EOFILE * eofile)
{
   struct host_entry * host;
   struct ip_hostentry * iphost = NULL;
   
   if(br_formhost(sess, &host, &iphost))
   {
      if(iphost)
      {
         maphost(sess, NULL, iphost, 0);
         return 0;
      }
      /* no info for passed host - possible Buster stopped &
       * restarted since the referrer page was displayed.
       */
      wi_printf(sess, "Computer information not found.<br>");
      return 1;
   }
   maphost(sess, host, NULL, 0);
   return 0;
}

/* ht_showip() - SSI routine.
 *
 * display a single Internet (IP) host.
 *
 * Returns 0 if OK. 
 */

int
ht_showip(wi_sess * sess, EOFILE * eofile)
{
   struct host_entry * host;
   struct ip_hostentry * iphost = NULL;
   
   br_formhost(sess, &host, &iphost);

   if(!iphost)
   {
      /* no info for passed host - possible Buster stopped &
       * restarted since the referrer page was displayed.
       */
      wi_printf(sess, "Computer information not found.<br>");
      return 1;
   }
   maphost(sess, NULL, iphost, 0);
   return 0;
}

/* static helpers for ht_connmap() Map drawing SSI routine */

char * vcnline1 = "<tr><td width=20%c>&nbsp;</td><td width=26%c>&nbsp;</td><td width=8%c align=center>"
                  "<table align=center cellspacing=1 >\n<tr>";

char * vcnline2 = "<td width=5 bgcolor=#ffffff title=\"Connections to Internet\" >"
      "<a href=\"connview.htm?index=%lu&ptr=%lu\" border=0 ><img src=\"%s\" "
      "border=0 ></a></td>\n";

char * vcnline3 = "</tr></table></td><td width=26%c>&nbsp;</td><td width=20%c>&nbsp;</td>\n</tr>";


/* br_vcnlines(struct httpd * hp,int layer)  -  helper routine for
 * ht_connmap() Map drawing SSI routine. Draws vertical connection
 * lines on a map page. "layer" parm is M_ISLOCAL, M_ISISP, or M_INET.
 * 
 * Returns 0 if OK, -1 to abort SSI.
 */

int
br_vcnlines(wi_sess * sess,
            int layer,              /* M_ISLOCAL, M_ISISP, or M_INET */
            struct congroup * cns,  /* array of connections */
            int conns)              /* entries in cns */
{
   int   i;
   int   index;      /* connection hash table index */
   struct conn_entry * conn;     /* tmp local copy */

   wi_printf(sess, vcnline1, '%', '%', '%');

   for(i = 0; i < conns; i++)
   {
      if(i >= MAXCONNS)
         break;

      switch(layer)
      {
         case M_ISLOCAL:     /* local net, show all */
            break;
         case M_INET:     /* Internet */
            if(cns[i].m_flags & M_ISLOCAL)
               continue;
            break;
         default:
            dtrap();    /* bad parameter */
            return -1;
      }
      /* Print with color, connection info */
      conn = cns[i].conn[0];
      index = HASH_CE(conn->rem.ce_addr, conn->loc.ce_addr, 
            conn->rem.ce_port, conn->loc.ce_port);
      wi_printf(sess, vcnline2, index, conn, br_vcimage(cns[i].port) );
   }
   wi_printf(sess, vcnline3, '%', '%');

   return 0;
}

/* next_cn() - Get next connection matching the passed flag from a list. The
 * passed index parm is incremented to the connection after the one returned. 
 *
 * If no cn is found with the flag then null is returned
 */

struct congroup *
next_cn(int * indexp, int max, struct congroup * cns, int flag)
{
   int   i;    /* local index */
   for(i = *indexp; i < max; i++)
   {
      if(cns[i].m_flags & flag)
      {
         *indexp = (i + 1);    /* indicate next index */
         return &cns[i];
      }
   }
   *indexp = max;
   return NULL;
}

/* name_gt(n1, n2) - 
 *
 * Case-insensitive alphabetical compare of two machine 
 * names (usually DNS). 
 *
 * Return 1 if n1 is greater than n2, 
 *    else return 0 if they are the same, and 
 *    else return -1 if n1 is less than n2.
 */

int
name_gt(char * n1, char * n2)
{
   u_char   c1, c2;

   while(*n1 > ' ')
   {
      c1 = (*n1++);
      if((c1 >= 'A') && (c1 <= 'Z'))   /* convert letter case */
         c1 &= 0x40;
      c2 = (*n2++);
      if((c2 >= 'A') && (c2 <= 'Z'))   /* convert letter case */
         c2 &= 0x40;
      if(c1 > c2)
         return 1;
      if(c1 < c2)
         return -1;
   }
   return 0;
}


/* br_sortcns() - sort connection group cns for display */

void
br_sortcns(struct congroup * cns, int count)
{
   struct congroup tmp;
   int   swaps;
   int   i;

   if(count < 2)     /* if 0 or 1 entry, nothing to do */
      return;

   for(;;)
   {
      swaps = 0;

      for(i = 0; i < count-1; i++)
      {
         int   gt;      /* TRUE if cns[i] > cns[i+1] */

         gt = FALSE;
         switch(br_cfg.sortparm)
         {
         case 'i':      /* sort by ipaddr */
            if(htonl(cns[i].ipaddr) > htonl(cns[i+1].ipaddr))
               gt = TRUE;
            break;
         case 't':      /* sort by traffic */
            if(cns[i].count > cns[i+1].count)
               gt = TRUE;
            break;
         case 'I':      /* sort by idle time */
            if(cns[i].conn[0]->ce_time > cns[i+1].conn[0]->ce_time)
               gt = TRUE;
            break;
         case 's':      /* sort by start time */
            if(cns[i].conn[0]->ce_start > cns[i+1].conn[0]->ce_start)
               gt = TRUE;
            break;
         case 'p':      /* sort by port */
            if(cns[i].port > cns[i+1].port)
               gt = TRUE;
            break;
         case 'n':      /* sort by name */
            {
               char * n1, *n2;
               struct ip_hostentry * ipe;
               
               /* Init each name to "z", then try to look up a name in 
                * the DNS system for the remnote host's IP address. If 
                * found, repklac ethe "z" with the real name.
                */
               n1 = "z";
               ipe = br_ipanyif(cns[i].ipaddr);
               if(ipe)
                  n1 = br_getname(ipe);

               n2 = "z";
               ipe = br_ipanyif(cns[i+1].ipaddr);
               if(ipe)
                  n2 = br_getname(ipe);

               if(name_gt(n1, n2) == 1)
                  gt = TRUE;
            }
            break;
         default:
            dtrap();    /* unknown sortparm? */
            break;
         }

         /* See if we need to swap the two items cns[i] and cns[i+1] */
         if(gt)
         {
            tmp = cns[i];
            cns[i] = cns[i+1];
            cns[i+1] = tmp;
            swaps++;
         }
      }
      if(swaps == 0)
         return;
   }
}


/* mapcon(hp, cn) - draw a horizontal connection line (or set of lines) 
 *
 * Helper routine for hostrow()
 *
 */

void
mapcon(wi_sess * sess, struct congroup * cn)
{
   char *   service;    /* Service for port, e.g. WWW or SMTP */
   char *   colorfile;  /* color for connection */
   char *   imgfile;    /* name of image file for endpoints */
   char *   status;
   int      size;       /* connection thickness, in pixels */
   int      i;          /* loop index */
   u_long   traffic;    /* swag at traffic rate */
   u_long   active;     /* time connection was active */
   int      index;      /* connection hash index */
   struct conn_entry * conn;  /* local copy of cn->conn */

   wi_printf(sess, "\n<table width=100%c border=0 cellpadding=0 cellspacing=1 ><tr>", '%');

   /* print the name (service type) of the connection */
   if(cn->conn[0]->ce_type == H_ICMP)
   {
      struct icmp * icmp;
      char * alarmtext = "";
      char * title = "name or service of connection";
 
      icmp = (struct icmp *)(cn->conn[0]->rem.app);
      if(icmp->type > MAX_ICMP_TYPE)
         service = icmp_types[BAD_ICMP_TYPE];
      else if((icmp->type == ECHOREQ) || (icmp->type == ECHOREP))
      {
         if(cn->conn[0]->ce_flags & CEF_LOCALSERVER)
         {
            service = "PING from Internet";
            alarmtext = "<font color=#FF0000><blink>";
            title = "This is often a worm or hacker looking for targets";
         }
         else
           service = "PING from local";
      }
      else
         service = icmp_types[icmp->type];

      wi_printf(sess, "<td align=center title=\"%s\" ><font size=-6 >%s", 
         title, alarmtext);
      wi_printf(sess, "ICMP: %s</td></tr>", service );
   }
   else if((service = br_portname(cn->port)) != NULL)
   {
      wi_printf(sess, "<td align=center><font size=-6>");
      wi_printf(sess, "%s</td></tr>", service);
   }
   else
   {
      wi_printf(sess, "<td align=center><font size=-6>");

      /* Try the local port of first connection. This may get a hit on 
       * Microsoft Netbeui probes from remote worms.
       */
      service = br_portname(cn->conn[0]->loc.ce_port);
      if(service)
         wi_printf(sess, "port %u->%s</td></tr>", cn->port, service);
      else
         wi_printf(sess, "port %u</td></tr>", cn->port);
   }

   colorfile = br_portimage(cn->port);

   /* Display horizontal connection line images. Loop through 
    * cn->conn[] array if there are multiple connections of this remote 
    * port/IP combo.
    */
   for(i = 0; i < cn->count; i++)
   {
      if(i >= MAXCONNS)     /* limit how many lines we display */
         break;
      conn = cn->conn[i];
      traffic = conn->rem.ce_outpkts + conn->loc.ce_outpkts;
      active = (conn->ce_time - conn->ce_start)/TPS;
      traffic /= (active|1);        /* packets/seconds */
      traffic = (int)pps_level((int)traffic);
      size = 4 + (traffic/4);       /* size range 4-8 */

      /* choose an image file & status text for this connection */
      if(conn->ce_flags & CEF_RESETREMOTE)
      {
         imgfile = "cnkilled.gif";
         status = "killed";
      }
      else if(conn->ce_type == H_TCP)  /* TCP connections */
      {
         if(conn->ce_state <= 4)       /* hardcode of TCPS_ESTABLISHED */
         {
            imgfile = NULL;            /* open connection */
            status = "active";
         }
         else
         {
            imgfile = "cnclose.gif";
            status = "closing";
         }
      }
      else     /* UDP, ICMP, or other... */
      {
         if(conn->ce_time > (cticks - (br_cfg.udp_idle * TPS)))
         {
            imgfile = "cnidle.gif";
            status = "active";
         }
         else
         {
            imgfile = NULL;   /* open connection */
            status = "idle";
         }
      }

      if((cn->port == 80) && (conn->data[0]))
      {
         /* We have a URL - print it in tooltip */
         wi_printf(sess, 
            "\n<tr><td height=%d align=center title=\"%s\">", 
            size, conn->data );
      }
      else
      {
         wi_printf(sess, 
            "\n<tr><td height=%d align=center title=\"%s connection - %s\">", 
            size, service?service:"data", status );
      }

      /* Get connection's hash index so we can build a link into
       * connview.htm page for connection detail
       */
      index = HASH_CE(conn->rem.ce_addr, conn->loc.ce_addr, 
               conn->rem.ce_port, conn->loc.ce_port);

      /* print the connection, with link, color, width based on traffic, and
       * an optional graphic based on the state.
       */
      wi_printf(sess, "<a href=\"connview.htm?index=%d&ptr=%lu\" border=0 >",
         index, conn);
      if(imgfile)
      {
         wi_printf(sess, "<img src='%s' height=%d width=33%c border=0 >", 
            colorfile, size, '%');
         wi_printf(sess, "<img src='%s' height=%d width=33%c border=0 >", 
            imgfile, size, '%');
         wi_printf(sess, "<img src='%s' height=%d width=33%c border=0 ></a></td>", 
            colorfile, size, '%');
      }
      else
      {
         wi_printf(sess, "<img src='%s' height=%d width=100%c border=0 ></a></td>", 
            colorfile, size, '%');
      }
   }

   wi_printf(sess, "</tr></table></td>\n" );
   return;
}


/* map_cnhost() - helper for ht_connmap(). draws a map entry for
 * the remote host in a cn struct.
 */

void
map_cnhost(wi_sess * sess, struct congroup * cn)
{
   struct ip_hostentry * iphost;

   iphost = br_ipanyif(cn->conn[0]->rem.ce_addr);
   if(iphost)
   {
      if((iphost->mac_host) && (iphost->mac_host->iphost == iphost))
      {
         maphost(sess, iphost->mac_host, NULL, 0);
      }
      else
      {
         /* See if we can assign an image to the host based on local port */
         if(cn->conn[0]->loc.ce_port == 0x87)   /* MS POPUP? */
         {
            if(iphost->imagefile[0] == 0)
               strcpy(iphost->imagefile, "mspopmsg.gif");  /* Gotcha! */
         }
         maphost(sess, NULL, iphost, 0);
      }
   }
   else
      maphost(sess, NULL, NULL, cn);

   return;
}


/* helper function - draws a connection map row with one or two hosts
 * of the type passed.
 */

static void
hostrow(wi_sess * sess,
        int htype,               /* M_ISLOCAL, M_ISISP, M_INET */
        struct congroup * cns, 
        int typecount,           /* number of cns entries of htype */
        int conns)               /* total number of cns entries */
{
   int   row = 0;
   int   cni = 0;
   struct congroup * cn;
   char * imgfile;


   /* print the first row of vertical connection bars */
   br_vcnlines(sess, htype, cns, conns);

   /* print two connections per row */
   for(row = 0; row <= (conns/2); row++)
   {
      cn = next_cn(&cni, conns, cns, htype);
      if(!cn)
         break;
      if(cn->port != cn->conn[0]->rem.ce_port)
      {
         dtrap();
         wi_printf(sess, "</td></tr>\n");
         return;
      }
      wi_printf(sess, "<tr><td align=center width=20%c>", '%');
      map_cnhost(sess, cn);
      wi_printf(sess, "</td><td width=26%c>\n", '%');
      mapcon(sess, cn);

      /* select and print image for center of network */
      switch(htype)
      {
      case M_ISLOCAL:
         imgfile = "lan1.gif";
         break;
      case M_INET:
         if(typecount <= 2)
            imgfile = "inet1.gif";
         else
         {
            if(row == 0)
               imgfile = "inettop.gif";
            else if(cni >= (conns - 1))
               imgfile = "inetbot.gif";
            else
               imgfile = "inetmid.gif";
         }
         break;
      default:
         dtrap();
         return;
      }

      wi_printf(sess, "</td><td align=center width=8%c  title=\"The Internet\"><img src=\"%s\" height=110 "
         "width=70 ></td>\n<td width=26%c >", '%', imgfile, '%');

      cn = next_cn(&cni, conns, cns, htype);

      if(cn)
      {
         if(cn->port != cn->conn[0]->rem.ce_port)
         {
            /* this may mean a bug, or it may happen because the conn 
             * structure was freed by the timer routine. If the latter,
             * the whole structure should be zeroed out.
             */
            dtrap();
            wi_printf(sess, "</td></tr>\n");
            return;
         }
         mapcon(sess, cn);
      }
      wi_printf(sess, "</td><td width=20%c>\n", '%');
      if(cn)
         map_cnhost(sess, cn);

      wi_printf(sess, "</td></tr>\n");
   }
   return;
}



/* ht_connmap() - SSI routine.
 * 
 * Draws a connection map for passed host
 *
 * Returns 0 if OK. 
 */

int
ht_connmap(wi_sess * sess, EOFILE * eofile)
{
   int      i,j;
   struct conn_entry * conn;
   struct ip_hostentry * iphost;
   struct host_entry * machost;
   struct congroup * cns;     /* array of cns entries, one per connection */
   int      cns_max;          /* number of entries in cns[] */    
   int      conns;            /* count of connections to display */
   int      totconns;         /* total connections for iphost */
   int      lanhosts = 0;
   int      inethosts = 0;

   /* Get the mac and IP host */
   i = br_formhost(sess, &machost, &iphost);
   if((i == -1) || (iphost == NULL))
      return -1;

   /* make a list of up to MAPCONNS connections, based on remote port and IP. Only
    * count each port/IP combo once.
    */
   conns = totconns = 0;         /* clear connection counts */
   if(iphost->conns == 0)        /* if no connections... */
   {
      wi_printf(sess, "<tr><td><center>"
         "No current network connections</td></tr>");
      return 0;
   }
   cns_max = iphost->conns + 5;     /* Allow for a few new connections during processing */
   cns = (struct congroup *)npalloc(sizeof(struct congroup) * cns_max);
   if(!cns)
   {
      dtrap();    /* was cns_max too big? */
      return -1;
   }

   /* We've now decided how many connectrions we will map, and  
    * tosome extent which ones. We don;t want any of these 
    * dissappearing in the code below, so set a db modules flag 
    * to prevent these deletions.
    */
   br_dbnodelete++;

   for(i = 0; i < HASH_CONN_ENTRIES; i++)
   {
      conn = ce_table[i];
      while(conn)
      {
         /* make sure connection involves this host */
         if((conn->loc.ce_addr != iphost->ipaddr) && 
            (conn->rem.ce_addr != iphost->ipaddr))
         {
            conn = conn->next;
            continue;
         }

#ifdef NPDEBUG
         /* Check for problem where local and remote addresses are 
          * swapped. -JB- 
          */
         if((conn->rem.ce_addr == iphost->ipaddr) &&     /* remote is us */
            ((conn->loc.ce_addr & nets[conn->ce_iface]->snmask) !=    /* local is remote */
              (nets[conn->ce_iface]->n_ipaddr & nets[conn->ce_iface]->snmask)))
         {
            if( ! br_isbcast(conn->ce_iface, conn->loc.ce_addr) )
            {
               struct endpoint   tmp;

               dtrap();    /* not allowed */
               /* Try swapping the ends? Fix later. -JB- */
               MEMCPY(&tmp, &conn->rem, sizeof(struct endpoint));
               MEMCPY(&conn->rem, &conn->loc, sizeof(struct endpoint));
               MEMCPY(&conn->loc, &tmp, sizeof(struct endpoint));

               /* Invert client-server flags */
               if(conn->ce_flags & CEF_LOCALCLIENT)
               {
                  conn->ce_flags |= CEF_LOCALSERVER;
                  conn->ce_flags &= ~CEF_LOCALCLIENT;
               }
               if(conn->ce_flags & CEF_LOCALSERVER)
               {
                  conn->ce_flags |= CEF_LOCALCLIENT;
                  conn->ce_flags &= ~CEF_LOCALSERVER;
               }

            }
         }
#endif   /* NPDEBUG */

         totconns++;    /* bump total connections count */

         /* if this session is hiding local traffic, skip local connections */
         if((br_cfg.hide_parm & HP_LOCAL) && islocal_ip(conn->ce_iface, conn->rem.ce_addr))
         {
            conn = conn->next;
            continue;
         }

         /* If flag is set, Filter out Buster Web GUI connections. */
         if(br_cfg.hide_parm & HP_BUSTER)
         {
            if((conn->loc.ce_addr == nets[conn->ce_iface]->n_ipaddr) &&
               (conn->loc.ce_port == htons(80)))
            {
               conn = conn->next;
               continue;
            }
            if((conn->rem.ce_addr == nets[conn->ce_iface]->n_ipaddr) &&
               (conn->rem.ce_port == htons(80)))
            {
               conn = conn->next;
               continue;
            }
         }

         /* scan the cns array to see if we already have this type */
         for(j = 0; j < cns_max; j++)
         {
            /* See if this conn matches this record port/IP combo */
            if((cns[j].port == conn->rem.ce_port) &&
               (cns[j].ipaddr == conn->rem.ce_addr))
            {
               int ct = cns[j].count;        /* Got a match */
               if(ct < MAXCONNS)             /* see if array has room */
                  cns[j].conn[ct] = conn;    /* save conn ptr */
               cns[j].count = ct + 1;        /* bump the count */
               break;
            }
         }
         if(j >= cns_max)    /* got a new port/IP combo? */
         {
            cns[conns].port = conn->rem.ce_port;
            cns[conns].ipaddr = conn->rem.ce_addr;
            cns[conns].conn[0] = conn;
            cns[conns].count = 1;

            /* If connection is on LAN, set flag */
            if(islocal_ip(conn->ce_iface, cns[conns].ipaddr))
            {
               cns[conns].m_flags |= M_ISLOCAL;
               lanhosts++;
            }
            else  /* default to internet host */
            {
               cns[conns].m_flags |= M_INET;
               inethosts++;
            }
            conns++;
            if(conns >= cns_max)
            {
               dtrap();    /* shouldn't happen anymore */
               goto conns_full;
            }
         }
         conn = conn->next;
      }
   }

conns_full:

   if(conns == 0)
   {
      wi_printf(sess, "<tr><td><center>"
         "No current network connections</td></tr>");
      npfree(cns);
      br_dbnodelete--;
      return 0;
   }

   /* sort cns for display */
   br_sortcns(cns, conns);

   if(lanhosts)
      hostrow(sess, M_ISLOCAL, cns, lanhosts, conns);
   if(inethosts)
      hostrow(sess, M_INET, cns, inethosts, conns);

   br_dbnodelete--;  /* release "no delete" flag */

   npfree(cns);
   return 0; /* OK return */
}

/* ht_recentlist() - SSI routine.
 * 
 * This is invoked by recentlist.ssi. It displays each of the blocking
 * events for the passed "host" in reverse chronological order.
 *
 * Returns 0 if OK. 
 */

int
ht_recentlist(wi_sess * sess, EOFILE * eofile)
{
   ip_addr  ipaddr;
   int      i;
   int      conns;
   char *   msg;                 /* error message from lib routine */
   struct conn_entry * conn;
   struct ip_hostentry * iphost;
   struct conn_entry ** clist;   /* list of blocked connections */
   int      clist_max;           /* number of entries in cns[] */

   /* get the IP address of the host */
   msg = wi_formipaddr(sess, "host", &ipaddr);
   if(msg)
   {
      wi_printf(sess, "<br>No info for this IP address<br>");
      dtrap();    /* Bad HTML/CGI ? */
      return 0;
   }
   iphost = br_ipanyif(ipaddr);

   /* make a list of up to MAPCONNS connections, based on remote port and IP. Only
    * count each port/IP combo once.
    */
   conns = 0;            /* clear connection counts */
   clist_max = iphost->conns;
   if(clist_max == 0)   /* no connections == no blocking - just return */
   {
      wi_printf(sess, "<tr class=\"blockingtint\" ><td colspan=6 >No recent blocking has occured.</td></tr>");
      return 0; /* OK return */
   }
   clist = (struct conn_entry **)npalloc(sizeof(char *) * clist_max);
   if(!clist)
   {
      dtrap();    /* was clist_max too big? */
      return -1;
   }

   /* loop through connection table, saving connections involving this
    * host that were blocked.
    */
   for(i = 0; i < HASH_CONN_ENTRIES; i++)
   {
      conn = ce_table[i];
      while(conn)
      {
         /* make sure connection involves this host */
         if((conn->loc.ce_addr != iphost->ipaddr) && 
            (conn->rem.ce_addr != iphost->ipaddr))
         {
            conn = conn->next;
            continue;
         }
         /* Only save pointers to blocked connections */
         if(conn->block.type == 0)
         {
            conn = conn->next;
            continue;
         }

         *(clist + conns) = conn;
         conns++;    /* bump total connections count */
         if(conns >= clist_max)
         {
            dtrap();    /* shouldn't happen */
            goto conns_full;
         }
         conn = conn->next;
      }
   }
conns_full:

   if(conns == 0)
   {
      wi_printf(sess, "<tr class=\"networktint\' ><td colspan=5 >No recent blocking has occured.</td></tr>");
      npfree(clist);
      return 0; /* OK return */
   }

   /* sort connections by last activity time. Newest values (most 
    * recent activity) go at the front of the list.
    */
   for(i = 0; i < conns; i++)
   {
      int   j;
      struct conn_entry * tmp;

      /* make sure clist[i] is lowest in remainder of list */
      for(j = i; j < conns; j++)
      {
         if(clist[i]->ce_time < clist[j]->ce_time)
         {
            tmp = clist[i];
            clist[i] = clist[j];
            clist[j] = tmp;
         }
      }
   }

   for(i = 0; i < conns; i++)
   {
      struct control_port *   cpt;
      struct b_path *         path;
      struct ctdesc *         ctl;
      struct control_ip *     ipe;
      char *   cssclass;
      char *   type;
      char *   filter;
      char     porttext[12];
      int      cn_index;

      conn = clist[i];

      /* collect the strings & data to print for this line */
      if(i & 1)
         cssclass = "blanktint";
      else
         cssclass = "networktint";

      /* Only one of these will be used. We set all three to avoid
       * Brain-dead compiler warnings.
       */
      ipe = (struct control_ip *)conn->block.ptr;
      path = (struct b_path *)conn->block.ptr;
      cpt = (struct control_port *)conn->block.ptr;

      switch(conn->block.type)
      {
      case CTF_TYPEIP:
         if(ipe->ci_name)
         {
            type = "DNS name";
            filter = ipe->ci_name->cu_name;
         }
         else
         {
            type = "IP address";
            filter = print_ipad(ipe->ci_ipaddr);
         }
         ctl = &ipe->ctl;
         break;
      case CTF_TYPEPATH:
         type = "URL path";
         filter = path->pathname;
         ctl = &path->ctl;
         break;
      case CTF_TYPEPORT:
         type = "Service (port)";
         filter = br_portname(cpt->cp_port);
         if(filter == NULL)
         {
            sprintf(porttext, "%d", cpt->cp_port);
            filter = porttext;
         }
         ctl = &cpt->ctl;
         break;
      default:
         dtrap();    /* bad blocking type? */
         npfree(clist);
         return -1;
      }

      /* Make sure offense description is not NULL. If it is, then
       * default to 1st entry in motives array .
       */
      if(ctl->offense == NULL)
      {
//         dtrap();    /* does this still happen? 7/15 - Yes. Shit. */
         ctl->offense = &motives[0];
      }

      /* Print out a line for this block event */
      wi_printf(sess, 
         "<tr class=\"%s\" ><td title=\"Type of filter which triggered the block\">\n"
         "%s</td><td title=\"Pattern which triggered the block\">\"%s\"</td>\n"
         "<td titlew=\"Motive for making the block filter\">%s</td>\n", 
         cssclass, type, filter, ctl->offense->desc);

      wi_printf(sess, "<td title=\"seconds since this block occured\">"
         "%lu <font size=-1>seconds</td>\n",
         (cticks - conn->ce_time)/TPS );

      cn_index = HASH_CE(conn->rem.ce_addr, conn->loc.ce_addr, 
         conn->rem.ce_port, conn->loc.ce_port);

      wi_printf(sess, "<td title=\"Open connection detail window\" "
         "class=\"buttons bnetwork\" >"
         "<a href=\"connview.htm?index=%ld&ptr=%ld\" >Details</a></td>",
            cn_index, conn);

      /* if the filter that killed this conn has been deleted,
       * then indicte this, otherwise make a button to delete it. 
       */
      if(conn->block.ptr == NULL)
      {
         wi_printf(sess, "<td>Deleted</td></tr>");
      }
      else     /* A filter exists */
      {
         int access;

         switch(conn->block.type)
         {
         case CTF_TYPEIP:
          {
            struct control_ip * ipe;
            ipe = (struct control_ip *)conn->block.ptr;

            access = ctl_getaccess(&ipe->ctl);
            if(access == 2)
            {
               wi_printf(sess, "<td>Timer</td></tr>");
            }
            else if(access == 1)
            {
               wi_printf(sess, "<td>Whitelist</td></tr>");
            }
            else
            {
               wi_printf(sess, "<td title=\"Click to disable the filter\" "
                  "class=\"buttons bnetwork\"  >\n"
                  "<a href=\"unblock.htm?host=%u.%u.%u.%u\">Unblock</a></td></tr>",
                  PASS_IPADDR(htonl(ipe->ci_ipaddr)) );
            }
            break;
          }
         case CTF_TYPEPATH:
          {
            struct b_path * bp;
            bp = (struct b_path *)conn->block.ptr;

            access = ctl_getaccess(&bp->ctl);
            if(access == 2)
            {
               wi_printf(sess, "<td>Timer</td></tr>");
            }
            else if(access == 1)
            {
               wi_printf(sess, "<td>Whitelist</td></tr>");
            }
            else
            {
               wi_printf(sess, "<td title=\"Click to disable the filter\" "
                  "class=\"buttons bnetwork\"  >\n"
                  "<a href=\"unblock.htm?path=%ld\">Unblock</a></td></tr>",
                  path);
            }
            break;
          }
         default:
            wi_printf(sess, "<td>Can't unblock</td></tr>");
         }
      }
   }

   npfree(clist);
   return 0; /* OK return */
}

/* ht_iflist() - SSI routine.
 *
 * Print a list of system interfaces with selection options.
 *
 * Returns 0 if OK. 
 */

int
ht_iflist(wi_sess * sess, EOFILE * eofile)
{
   int iface;        /* loop index */
   int ifcheck;      /* which iface should be "checked" */
   char * disabled;  /* string for optional 'disable' info */

   if(br_autoifview)
      ifcheck = -1;
   else
      ifcheck = br_ifview;

   wi_printf(sess, "<tr><td><input type=radio name=\"view\" value=\"Auto\" %s ></td>",
      (ifcheck == -1)?"CHECKED":"" );

   wi_printf(sess, "<td>Automatically switch to busiest network<br></td><td></td></tr>");

   for(iface = 0; iface < (int)ifNumber; iface++)
   {
      if((nets[iface]->ifOperStatus != 1) ||
         (nets[iface]->ifAdminStatus != 1))
      {
         disabled = "(disabled)";
      }
      else
      {
         disabled = "";
      }

      wi_printf(sess, "<tr><td><input type=radio name=\"view\" value=\"%d\" %s ></td>\n", 
         iface, iface == ifcheck?"CHECKED":" " );
      wi_printf(sess, "<td>Always %s<br></td>\n", 
         nets[iface]->description );
      wi_printf(sess, "<td class=\"%s\"title=\"Level of recent activity (0 is lowest)\" >"
         "%lu %s</td></tr>\n",
         act_color(1, TPS, nets[iface]->traffic), 
         nets[iface]->traffic, 
         disabled);
   }
	return 0; /* OK return */
}


/* viewnet_cgi() - CGI routine.
 *
 * Sets the newwork interface for viewing.
 *
 */

char *
viewnet_cgi(wi_sess * sess, EOFILE * eofile)
{
   char *	view;		/* to be set from form */
   int      newif;

	view = wi_formvalue(sess, "view");

   if(view)
   {
      if(*view == 'A')     /* "Auto" select */
      {
         br_autoifview = TRUE;
         /* Save the new auto-select flag in the NV system. */
         br_saveparms();
      }
      else if(isdigit(*view) == FALSE)
      {
         dtrap();
         return("HTML error - bad view iface");
      }
      else
      {
         br_autoifview = FALSE;
         newif = atoi(view);

         /* Set new iface and seve to NV system */
         br_setiface(&br_ifview, newif);
      }
   }
   else
   {
      dtrap();
      return "No view option";
   }

   br_autoselect();
   
   wi_redirect(sess, "netmap.htm");
   return NULL;
}

/* del_hide_cgi() - CGI routine.
 *
 * Either hide or delete the MAC host passed.
 *
 * Returns bitmask (see manual).
 */

char *
del_hide_cgi(wi_sess * sess, EOFILE * eofile)
{
   char *   hosttext;
   u_char   macaddr[6];
   struct host_entry * machost;

   hosttext = wi_formvalue(sess, "host");
   if(hosttext == NULL)
      return("del_hide: missing MAC");

   if(parse_mac(hosttext, macaddr) != 0)
      return("del_hide: bad MAC");

   machost = br_hostlookup((char*)macaddr, br_ifview);
   if(!machost)
      return("del_hide: bad MAC");

   host_delete(machost, HASH_HOST(macaddr) );

   wi_redirect(sess, "netmap.htm");
   return NULL;
}


/* routine to select an image file for a passed MAC host */

char *
br_image(struct host_entry * host)
{
   char *   image = "unknown.gif";
   struct image * img;
   struct btnode * btnp;

   if(!host)
   {
      dtrap(); /* bug */
      return image;
   }
   /* if host has an assigned image; use that */
   if(host->imagefile[0] != 0)
      return (host->imagefile);

   /* if host is the router, use router image */
   if(host == brstates[host->iface].br_router)
      return "rte.jpg";

   if(host->h_flags & H_BCAST)
      return "bcast.gif";

   /* try MAC address lookup. If we get a hit, replace the unknown with it. */
   btnp = bt_access((struct btnode *)image_tree, 
      (u_long)(&host->macaddr[0]), bt_mactest );

   if(btnp)
   {
      img = (struct image *)btnp->data;
      image = img->imagefile;
   }
   else
   {
      if(host->h_flags & H_WINDOWS)
         image = "pc.gif";
   }
   if(host->h_flags & H_LINUX)
      image = "linux.gif";

   return image;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

