/*
 * FILENAME: except.c
 *
 * Copyright (c) 1998 Bruce Dawson
 *
 * This source file contains the exception handler for recording error
 * information after crashes. See ExceptionHandler.h for information
 * on how to hook it in.
 *
 * ROUTINES: lstrrchr (), hflush(), hprintf(), 
 * ROUTINES: FormatTime(), module (), DumpModuleInfo(), modules (), 
 * ROUTINES: DumpModuleList(), DumpSystemInformation(), 
 * ROUTINES: GetExceptionDescription(), GetFilePart(), DumpStack(), 
 * ROUTINES: DumpRegisters(), RecordExceptionInfo(), 
 *
 ****> External entry point: RecordExceptionInfo();
 *
 *
 * Original Author:       Bruce Dawson
 *               brucedawson@cygnus-software.com
 *
 * Modified by:  Hans Dietrich
 *               hdietrich2@hotmail.com
 *
 * Version 1.1:  - reformatted output for XP-like error report
 *               - added ascii output to stack dump
 *
 * A paper by the original author can be found at:
 *     http: //www.cygnus-software.com/papers/release_debugging.html
 *
 */


/* Disable warnings generated by the Windows header files. */
#pragma warning(disable : 4514)
#pragma warning(disable : 4201)
#pragma warning(disable : 4115)

#define  _WIN32_WINDOWS 0x0500   /* for   IsDebuggerPresent */


#include "windows.h"
#include <tchar.h>

#ifndef _countof
#define  _countof(array)   (sizeof(array)/sizeof(array[0]))
#endif

const int NumCodeBytes = 16;     /* Number of code bytes to record. */
const int MaxStackDump = 3072;   /* Maximum number of DWORDS in stack dumps. */
const int StackColumns = 4;      /* Number of columns in stack dump. */

#define  ONEK     1024
#define  SIXTYFOURK  (64*ONEK)
#define  ONEM     (ONEK*ONEK)
#define  ONEG     (ONEK*ONEK*ONEK)

extern OSVERSIONINFO  osVer;         /* for GetVersionEx() */ 


#define BUSTER 1

#ifdef BUSTER
char * XCRASHREPORT_ERROR_LOG_FILE = "dumplog.txt";
/* externals - these should be defined and set outside of this module. */
extern   char     homedir[];
extern   char *   br_version;
extern   char *   panicmsg;      /* Set if panic was invoked */
extern   u_long   br_versnum;
#endif /* BUSTER */


/* FUNCTION: lstrrchr()
 * 
 * PARAM1: LPCTSTR string
 * PARAM2: int ch
 *
 * RETURNS: 
 */

static   TCHAR *
lstrrchr(LPCTSTR string, int ch)
{
   TCHAR *start = (TCHAR *)string;

   while (*string++)                        /* find end of string */
      ;
   /* search towards front */
   while (--string != start && *string != (TCHAR) ch)
      ;

   if (*string == (TCHAR) ch)                  /* char found ? */
      return (TCHAR *)string;

   return NULL;
}



#define  HPRINTF_BUFFER_SIZE  (8*1024)    /* must  be at least 2048 */
static TCHAR hprintf_buffer[HPRINTF_BUFFER_SIZE]; /* wvsprintf never prints more than one K. */
static int  hprintf_index = 0;



/* FUNCTION: hflush()
 * 
 * hprintf behaves similarly to printf, with a few vital differences.
 * It uses wvsprintf to do the formatting, which is a system routine,
 * *  thus avoiding C run time interactions. For similar reasons it
 * uses WriteFile rather than fwrite.
 * The one limitation that this imposes is that wvsprintf, and
 therefore hprintf, cannot handle floating point numbers.

 Too many calls to WriteFile can take a long time, causing
 confusing delays when programs crash. Therefore I implemented
 a simple buffering scheme for hprintf
 * PARAM1: HANDLE LogFile
 *
 * RETURNS: 
 */

static   void
hflush(HANDLE LogFile)
{
   if (hprintf_index > 0)
   {
      DWORD NumBytes;
      WriteFile(LogFile, hprintf_buffer, lstrlen(hprintf_buffer), &NumBytes, 0);
      hprintf_index = 0;
   }
}



/* FUNCTION: hprintf()
 * 
 * PARAM1: HANDLE LogFile
 * PARAM2: LPCTSTR Format
 * PARAM3: ...
 *
 * RETURNS: 
 */

static   void
hprintf(HANDLE LogFile, LPCTSTR Format, ...)
{
   va_list arglist;

   if (hprintf_index > (HPRINTF_BUFFER_SIZE-1024))
   {
      DWORD NumBytes;
      WriteFile(LogFile, hprintf_buffer, lstrlen(hprintf_buffer), &NumBytes, 0);
      hprintf_index = 0;
   }

   va_start( arglist, Format);
   hprintf_index += wvsprintf(&hprintf_buffer[hprintf_index], Format, arglist);
   va_end( arglist);
}


/* FUNCTION: FormatTime()
 * 
 * Format the specified FILETIME to output in a human readable format,
 * without using the C run time.
 * PARAM1: LPTSTR output
 * PARAM2: FILETIME TimeToPrint
 *
 * RETURNS: 
 */

static   void
FormatTime(LPTSTR output, FILETIME TimeToPrint)
{
   WORD Date, Time;

   output[0] = _T('\0');
   if (FileTimeToLocalFileTime(&TimeToPrint, &TimeToPrint) &&
       FileTimeToDosDateTime(&TimeToPrint, &Date, &Time))
   {
      wsprintf(output, _T("%d/%d/%d %02d:%02d:%02d"),
       (Date / 32) & 15, Date & 31, (Date / 512) + 1980,
       (Time >> 11), (Time >> 5) & 0x3F, (Time & 0x1F) * 2);
   }
}


/* FUNCTION: DumpModuleInfo()
 * 
 * Print information about a  code  module (DLL or EXE) such as its size, 
 * location, time stamp, etc.
 *
 * PARAM1: HANDLE LogFile
 * PARAM2: HINSTANCE ModuleHandle
 * PARAM3: int nModuleNo
 *
 * RETURNS: 
 */


int
DumpModuleInfo(HANDLE LogFile, HINSTANCE ModuleHandle, int nModuleNo)
{
   int   rc = FALSE;
   TCHAR szModName[MAX_PATH*2];
   IMAGE_NT_HEADERS * NTHeader;
   IMAGE_DOS_HEADER * DosHeader;
   HANDLE ModuleFile;
   TCHAR TimeBuffer[100];
   DWORD FileSize = 0;
   FILETIME LastWriteTime;

   ZeroMemory(szModName, sizeof(szModName));

   __try
   {
      if (GetModuleFileName(ModuleHandle, szModName, sizeof(szModName)-2) > 0)
      {
         /* If GetModuleFileName returns greater than zero then this must
          * be a valid code module address. Therefore we can try to walk
          * our way through its structures to find the link time stamp.
          */
         DosHeader = (IMAGE_DOS_HEADER*)ModuleHandle;
         if (IMAGE_DOS_SIGNATURE != DosHeader->e_magic)
            return FALSE;

         NTHeader = (IMAGE_NT_HEADERS*)((TCHAR *)DosHeader
          + DosHeader->e_lfanew);
         if (IMAGE_NT_SIGNATURE != NTHeader->Signature)
            return FALSE;

         /* open the code module file so that we can get its file date and size */
         ModuleFile = CreateFile(szModName, GENERIC_READ,
            FILE_SHARE_READ, 0, OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL, 0);

         TimeBuffer[0] = _T('\0');

         if (ModuleFile != INVALID_HANDLE_VALUE)
         {
            FileSize = GetFileSize(ModuleFile, 0);
            if (GetFileTime(ModuleFile, 0, 0, &LastWriteTime))
            {
               FormatTime(TimeBuffer, LastWriteTime);
            }
            CloseHandle(ModuleFile);
         }
         hprintf(LogFile, _T("Module %d\r\n"), nModuleNo);
         hprintf(LogFile, _T("%s\r\n"), szModName);

         hprintf(LogFile, _T("Image Base: 0x%08x  Image Size: 0x%08x\r\n"), 
            NTHeader->OptionalHeader.ImageBase, 
            NTHeader->OptionalHeader.SizeOfImage);

         hprintf(LogFile, _T("Checksum:   0x%08x  Time Stamp: 0x%08x\r\n"), 
            NTHeader->OptionalHeader.CheckSum,
            NTHeader->FileHeader.TimeDateStamp);

         hprintf(LogFile, _T("File Size:  %-10d  File Time:  %s\r\n"),
            FileSize, TimeBuffer);

         hprintf(LogFile, _T("\r\n"));

         rc = TRUE;
      }
   }
   /* Handle any exceptions by continuing from this point. */
   __except(EXCEPTION_EXECUTE_HANDLER)
   {
   }
   return rc;
}


/* FUNCTION: DumpModuleList()
 * 
 * Scan  memory   looking  for   code  modules (DLLs or EXEs). VirtualQuery is used
 * to find all the blocks of address space that were reserved or committed,
 * and ShowModuleInfo will display module information if they are code
 * modules.
 *
 * PARAM1: HANDLE LogFile
 *
 * RETURNS: 
 */

static   void
DumpModuleList(HANDLE LogFile)
{
   SYSTEM_INFO    SystemInfo;
   size_t   PageSize;
   int      nModuleNo   =  1;
   void *   LastAllocationBase   =  0;
   size_t   NumPages;
   size_t   pageNum = 0;

   GetSystemInfo(&SystemInfo);

   /* Set NumPages to the number of pages in the 4GByte address space,
    * while being careful to avoid overflowing ints
    */
   PageSize = SystemInfo.dwPageSize;
   NumPages = 4 * (size_t)(ONEG / PageSize);


   while (pageNum < NumPages)
   {
      MEMORY_BASIC_INFORMATION MemInfo;

      if (VirtualQuery((void *)(pageNum * PageSize), &MemInfo,
          sizeof(MemInfo)))
      {
         if (MemInfo.RegionSize > 0)
         {
            /* Adjust the page number to skip over this block of memory */
            pageNum += MemInfo.RegionSize / PageSize;
            if (MemInfo.State == MEM_COMMIT && MemInfo.AllocationBase >
                LastAllocationBase)
            {
               /* Look for new blocks of committed memory, and try
                * recording their module names - this will fail
                * gracefully if they aren't code modules
                */
               LastAllocationBase = MemInfo.AllocationBase;
               if (DumpModuleInfo(LogFile, 
                   (HINSTANCE)LastAllocationBase, 
                   nModuleNo))
               {
                  nModuleNo++;
               }
            }
         }
         else
            pageNum += SIXTYFOURK / PageSize;
      }
      else
         pageNum += SIXTYFOURK / PageSize;

      /* If VirtualQuery fails we advance by 64K because that is the
       * granularity of address space doled out by VirtualAlloc()
       */
   }
}



/* FUNCTION: DumpSystemInformation()
 * 
 * DumpSystemInformation
 * Record information about the user's system, such as processor type, amount
 * of memory, etc.
 *
 * PARAM1: HANDLE LogFile
 *
 * RETURNS: 
 */


static   void
DumpSystemInformation(HANDLE LogFile)
{
   FILETIME CurrentTime;
   TCHAR szTimeBuffer[100];
   TCHAR szModuleName[MAX_PATH*2];
   TCHAR szUserName[200];
   DWORD UserNameSize;
   SYSTEM_INFO SystemInfo;
   MEMORYSTATUS MemInfo;


   GetSystemTimeAsFileTime(&CurrentTime);
   FormatTime(szTimeBuffer, CurrentTime);

   hprintf(LogFile, _T("Error occurred at %s.\r\n"), szTimeBuffer);

   ZeroMemory(szModuleName, sizeof(szModuleName));
   if (GetModuleFileName(0, szModuleName, _countof(szModuleName)-2) <= 0)
      lstrcpy(szModuleName, _T("Unknown"));

   ZeroMemory(szUserName, sizeof(szUserName));
   UserNameSize = _countof(szUserName)-2;
   if (!GetUserName(szUserName, &UserNameSize))
      lstrcpy(szUserName, _T("Unknown"));

   hprintf(LogFile, _T("%s, run by %s.\r\n"), szModuleName, szUserName);

   /* print out operating system */
   hprintf(LogFile, _T("Operating system:  %d.%d.\r\n"), 
      osVer.dwMajorVersion, osVer.dwMinorVersion );

   GetSystemInfo(&SystemInfo);
   hprintf(LogFile, _T("%d processor(s), type %d.\r\n"),
    SystemInfo.dwNumberOfProcessors, SystemInfo.dwProcessorType);

   MemInfo.dwLength = sizeof(MemInfo);
   GlobalMemoryStatus(&MemInfo);

   /* Print out info on memory, rounded up. */
   hprintf(LogFile, _T("%d%% memory in use.\r\n"), MemInfo.dwMemoryLoad);
   hprintf(LogFile, _T("%d MBytes physical memory.\r\n"), (MemInfo.dwTotalPhys +
      ONEM - 1) / ONEM);
   hprintf(LogFile, _T("%d MBytes physical memory free.\r\n"), 
      (MemInfo.dwAvailPhys + ONEM - 1) / ONEM);
   hprintf(LogFile, _T("%d MBytes paging file.\r\n"), (MemInfo.dwTotalPageFile +
      ONEM - 1) / ONEM);
   hprintf(LogFile, _T("%d MBytes paging file free.\r\n"), 
      (MemInfo.dwAvailPageFile + ONEM - 1) / ONEM);
   hprintf(LogFile, _T("%d MBytes user address space.\r\n"), 
      (MemInfo.dwTotalVirtual + ONEM - 1) / ONEM);
   hprintf(LogFile, _T("%d MBytes user address space free.\r\n"), 
      (MemInfo.dwAvailVirtual + ONEM - 1) / ONEM);

   return;
}



/* FUNCTION: GetExceptionDescription()
 * 
 * Translate the exception code into something human readable
 *
 * PARAM1: DWORD ExceptionCode
 *
 * RETURNS: 
 */

struct ExceptionNames
{
    DWORD ExceptionCode;
    TCHAR *  ExceptionName;
};


#if 0  /* from winnt.h */
#define  STATUS_WAIT_0                    ((DWORD     )0x00000000L)    
#define  STATUS_ABANDONED_WAIT_0          ((DWORD     )0x00000080L)    
#define  STATUS_USER_APC                  ((DWORD     )0x000000C0L)    
#define  STATUS_TIMEOUT                   ((DWORD     )0x00000102L)    
#define  STATUS_PENDING                   ((DWORD     )0x00000103L)    
#define  STATUS_SEGMENT_NOTIFICATION      ((DWORD     )0x40000005L)    
#define  STATUS_GUARD_PAGE_VIOLATION      ((DWORD     )0x80000001L)    
#define  STATUS_DATATYPE_MISALIGNMENT     ((DWORD     )0x80000002L)    
#define  STATUS_BREAKPOINT                ((DWORD     )0x80000003L)    
#define  STATUS_SINGLE_STEP               ((DWORD     )0x80000004L)    
#define  STATUS_ACCESS_VIOLATION          ((DWORD     )0xC0000005L)    
#define  STATUS_IN_PAGE_ERROR             ((DWORD     )0xC0000006L)    
#define  STATUS_INVALID_HANDLE            ((DWORD     )0xC0000008L)    
#define  STATUS_NO_MEMORY                 ((DWORD     )0xC0000017L)    
#define  STATUS_ILLEGAL_INSTRUCTION       ((DWORD     )0xC000001DL)    
#define  STATUS_NONCONTINUABLE_EXCEPTION  ((DWORD     )0xC0000025L)    
#define  STATUS_INVALID_DISPOSITION       ((DWORD     )0xC0000026L)    
#define  STATUS_ARRAY_BOUNDS_EXCEEDED     ((DWORD     )0xC000008CL)    
#define  STATUS_FLOAT_DENORMAL_OPERAND    ((DWORD     )0xC000008DL)    
#define  STATUS_FLOAT_DIVIDE_BY_ZERO      ((DWORD     )0xC000008EL)    
#define  STATUS_FLOAT_INEXACT_RESULT      ((DWORD     )0xC000008FL)    
#define  STATUS_FLOAT_INVALID_OPERATION   ((DWORD     )0xC0000090L)    
#define  STATUS_FLOAT_OVERFLOW            ((DWORD     )0xC0000091L)    
#define  STATUS_FLOAT_STACK_CHECK         ((DWORD     )0xC0000092L)    
#define  STATUS_FLOAT_UNDERFLOW           ((DWORD     )0xC0000093L)    
#define  STATUS_INTEGER_DIVIDE_BY_ZERO    ((DWORD     )0xC0000094L)    
#define  STATUS_INTEGER_OVERFLOW          ((DWORD     )0xC0000095L)    
#define  STATUS_PRIVILEGED_INSTRUCTION    ((DWORD     )0xC0000096L)    
#define  STATUS_STACK_OVERFLOW            ((DWORD     )0xC00000FDL)    
#define  STATUS_CONTROL_C_EXIT            ((DWORD     )0xC000013AL)    
#define  STATUS_FLOAT_MULTIPLE_FAULTS     ((DWORD     )0xC00002B4L)    
#define  STATUS_FLOAT_MULTIPLE_TRAPS      ((DWORD     )0xC00002B5L)    
#define  STATUS_ILLEGAL_VLM_REFERENCE     ((DWORD     )0xC00002C0L)     
#endif

struct ExceptionNames ExceptionMap[] =
{
      {0x40010005, _T("a Control-C")},
      {0x40010008, _T("a Control-Break")},
      {0x80000002, _T("a Datatype Misalignment")},
      {0x80000003, _T("a Breakpoint")},
      {0xc0000005, _T("an Access Violation")},
      {0xc0000006, _T("an In Page Error")},
      {0xc0000008, _T("an invalid handle")},
      {0xc0000017, _T("a No Memory")},
      {0xc000001d, _T("an Illegal Instruction")},
      {0xc0000025, _T("a Noncontinuable Exception")},
      {0xc0000026, _T("an Invalid Disposition")},
      {0xc000008c, _T("a Array Bounds Exceeded")},
      {0xc000008d, _T("a Float Denormal Operand")},
      {0xc000008e, _T("a Float Divide by Zero")},
      {0xc000008f, _T("a Float Inexact Result")},
      {0xc0000090, _T("a Float Invalid Operation")},
      {0xc0000091, _T("a Float Overflow")},
      {0xc0000092, _T("a Float Stack Check")},
      {0xc0000093, _T("a Float Underflow")},
      {0xc0000094, _T("an Integer Divide by Zero")},
      {0xc0000095, _T("an Integer Overflow")},
      {0xc0000096, _T("a Privileged Instruction")},
      {0xc00000fD, _T("a Stack Overflow")},
      {0xc0000142, _T("a DLL Initialization Failed")},
      {0xe06d7363, _T("a Microsoft C++ Exception")},
};

static   const TCHAR *
GetExceptionDescription(DWORD ExceptionCode)
{
   int   i;

   for (i = 0; i < sizeof(ExceptionMap) / sizeof(ExceptionMap[0]); i++)
   {
      if (ExceptionCode == ExceptionMap[i].ExceptionCode)
         return ExceptionMap[i].ExceptionName;
   }

   return _T("an Unknown exception type");
}


/* FUNCTION: GetFilePart()
 * 
 * PARAM1: LPCTSTR source
 *
 * RETURNS: 
 */

static   TCHAR *
GetFilePart(LPCTSTR source)
{
   TCHAR *result = lstrrchr(source, _T('\\'));
   if (result)
      result++;
   else
      result = (TCHAR *)source;
   return result;
}

/* FUNCTION: DumpStack()
 * 
 * PARAM1: HANDLE LogFile
 * PARAM2: DWORD *pStack
 *
 * RETURNS: 
 */

static   void
DumpStack(HANDLE LogFile, DWORD * pStack)
{
   int      Count =  0;
   int      nDwordsPrinted =  0;
   int      i;
   int      n;
   int      j;
   char     c;
   DWORD *  pStackStart = pStack;

   hprintf(LogFile, _T("\r\n\r\nStack:\r\n"));

   __try
   {
      /* Esp contains the bottom of the stack, or at least the bottom of
       * the currently used area.
       */
      DWORD* pStackTop;

      __asm
      {
         /* Load the top (highest address) of the stack from the
          * thread information block. It will be found there in
          * Win9x and Windows NT.
          */
         mov eax, fs:[4]
         mov pStackTop, eax
      }

      if (pStackTop > pStack + MaxStackDump)
         pStackTop = pStack + MaxStackDump;


      while (pStack + 1 <= pStackTop)
      {
         if ((Count % StackColumns) == 0)
         {
            pStackStart = pStack;
            nDwordsPrinted = 0;
            hprintf(LogFile, _T("0x%08x: "), pStack);
         }

         if ((++Count % StackColumns) == 0 || pStack + 2 > pStackTop)
         {
            hprintf(LogFile, _T("%08x "), *pStack);
            nDwordsPrinted++;

            n  =  nDwordsPrinted;
            while (n < 4)
            {
               hprintf(LogFile, _T("         "));
               n++;
            }

            for (i = 0; i < nDwordsPrinted; i++)
            {
               DWORD dwStack = *pStackStart;
               for (j = 0; j < 4; j++)
               {
                  c  =  (char)(dwStack &  0xFF);
                  if (c < 0x20 || c > 0x7E)
                     c = '.';
#ifdef _UNICODE
                  {
                     WCHAR    w;
                     w = (WCHAR)c;
                     hprintf(LogFile, _T("%c"), w);
                  }
#else
                  hprintf(LogFile, _T("%c"), c);
#endif
                  dwStack = dwStack >> 8;
               }
               pStackStart++;
            }

            hprintf(LogFile, _T("\r\n"));
         }
         else
         {
            hprintf(LogFile, _T("%08x "), *pStack);
            nDwordsPrinted++;
         }
         pStack++;
      }
      hprintf(LogFile, _T("\r\n"));
   }
   __except(EXCEPTION_EXECUTE_HANDLER)
   {
      hprintf(LogFile, _T("Exception encountered during stack dump.\r\n"));
   }
}


/* FUNCTION: DumpRegisters()
 * 
 * PARAM1: HANDLE LogFile
 * PARAM2: PCONTEXT Context
 *
 * RETURNS: 
 */

static   void
DumpRegisters(HANDLE LogFile, PCONTEXT Context)
{
   /* Print out the register values in an XP error window compatible format. */
   hprintf(LogFile, _T("\r\n"));
   hprintf(LogFile, _T("Context:\r\n"));
   hprintf(LogFile, _T("EDI:    0x%08x  ESI: 0x%08x  EAX:   0x%08x\r\n"),
    Context->Edi, Context->Esi, Context->Eax);
   hprintf(LogFile, _T("EBX:    0x%08x  ECX: 0x%08x  EDX:   0x%08x\r\n"),
    Context->Ebx, Context->Ecx, Context->Edx);
   hprintf(LogFile, _T("EIP:    0x%08x  EBP: 0x%08x  SegCs: 0x%08x\r\n"),
    Context->Eip, Context->Ebp, Context->SegCs);
   hprintf(LogFile, _T("EFlags: 0x%08x  ESP: 0x%08x  SegSs: 0x%08x\r\n"),
    Context->EFlags, Context->Esp, Context->SegSs);
}


/* FUNCTION: RecordExceptionInfo()
 * 
 * PARAM1: PEXCEPTION_POINTERS pExceptPtrs
 * PARAM2: LPCTSTR lpszMessage
 *
 * RETURNS: 
 */

int      __cdecl
RecordExceptionInfo(PEXCEPTION_POINTERS pExceptPtrs, 
    LPCTSTR  lpszMessage)
{
   static int  bFirstTime = TRUE;
   TCHAR       szModuleName[MAX_PATH*2];
   TCHAR       szFileName[MAX_PATH*2];
   TCHAR *     pszCrashModuleFileName;
   HANDLE      hLogFile;
   PCONTEXT    Context   = pExceptPtrs->ContextRecord;
   TCHAR       szCrashModulePathName[MAX_PATH*2];
   MEMORY_BASIC_INFORMATION MemInfo;
   PEXCEPTION_RECORD Exception = pExceptPtrs->ExceptionRecord;
   BYTE *      code;
   DWORD *     pStack;
   DWORD *     pFrame;
   int         codebyte;
   char *      pszFilePart;


   if (!bFirstTime) /* Going recursive! That must mean this routine crashed! */
      return EXCEPTION_CONTINUE_SEARCH;
   bFirstTime = FALSE;

   ZeroMemory(szModuleName, sizeof(szModuleName));

   /* Create a filename to record the error information to.
    * Storing it in the executable directory works well.
    */
   if (GetModuleFileName(0, szModuleName, _countof(szModuleName)-2) <= 0)
      lstrcpy(szModuleName, _T("Unknown"));

   pszFilePart = GetFilePart(szModuleName);

   /* Extract the file name portion and remove it's file extension */
   lstrcpy(szFileName, pszFilePart);

#ifndef BUSTER
   lastperiod = lstrrchr(szFileName, _T('.'));
   if (lastperiod)
      lastperiod[0] = 0;

   /* Replace the executable filename with our error log file name */
   lstrcpy(pszFilePart, XCRASHREPORT_ERROR_LOG_FILE);
#else

   strncpy(szModuleName, homedir, MAX_PATH);
   strcat(szModuleName, "\\");
   strcat(szModuleName, XCRASHREPORT_ERROR_LOG_FILE);

#endif   /* BUSTER */

   hLogFile = CreateFile(szModuleName, GENERIC_WRITE, 0, 0,
      CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_WRITE_THROUGH, 0);

   if (hLogFile == INVALID_HANDLE_VALUE)
   {
      OutputDebugString(_T("Error creating exception report\r\n"));
      return EXCEPTION_CONTINUE_SEARCH;
   }

   /* Append to the error log */
   SetFilePointer(hLogFile, 0, 0, FILE_END);

   /* Print out a blank line to separate this error log from any previous ones
    * hprintf(hLogFile, _T("\r\n"));
    */

   Exception = pExceptPtrs->ExceptionRecord;
   Context   = pExceptPtrs->ContextRecord;

   szCrashModulePathName[MAX_PATH*2];
   ZeroMemory(szCrashModulePathName, sizeof(szCrashModulePathName));

   pszCrashModuleFileName = _T("Unknown");

   /* VirtualQuery can be used to get the allocation base associated with a
    * code address, which is the same as the ModuleHandle. This can be used
    * to get the filename of the module that the crash happened in.
    */
   if (VirtualQuery((void*)Context->Eip, &MemInfo, sizeof(MemInfo)) &&
       (GetModuleFileName((HINSTANCE)MemInfo.AllocationBase,
       szCrashModulePathName,
       sizeof(szCrashModulePathName)-2) > 0))
   {
      pszCrashModuleFileName = GetFilePart(szCrashModulePathName);
   }

   /* Print out the beginning of the error log in a Win95 error window
    * compatible format.
    */
   hprintf(hLogFile, _T("%s caused %s (0x%08x) exception \r\nin module %s at %04x:%08x.\r\n\r\n"),
       szFileName, GetExceptionDescription(Exception->ExceptionCode),
       Exception->ExceptionCode,
       pszCrashModuleFileName, Context->SegCs, Context->Eip);

   hprintf(hLogFile, _T("Exception handler called in %s.\r\n"), lpszMessage);

   hprintf(hLogFile, _T("Buster code version %s.\r\n"), br_version );

   if(panicmsg)
      hprintf(hLogFile, _T("panic message: %s\n"), panicmsg);

   DumpSystemInformation(hLogFile);

   /* If the exception was an access violation, print out some additional 
    * information, to the error log and the debugger.
    */
   if (Exception->ExceptionCode == STATUS_ACCESS_VIOLATION &&
       Exception->NumberParameters >= 2)
   {
      TCHAR szDebugMessage[1000];
      const TCHAR* readwrite = _T("Read from");
      if (Exception->ExceptionInformation[0])
         readwrite = _T("Write to");
      wsprintf(szDebugMessage, _T("%s location %08x caused an access violation.\r\n"),
       readwrite, Exception->ExceptionInformation[1]);

#ifdef _DEBUG
      /* The Visual C++ debugger doesn't actually tell you whether a read
       * or a write caused the access violation, nor does it tell what
       * address was being read or written. So I fixed that.
       */
      OutputDebugString(_T("Exception handler: "));
      OutputDebugString(szDebugMessage);
#endif

      hprintf(hLogFile, _T("%s"), szDebugMessage);
   }

   DumpRegisters(hLogFile, Context);

   /* Print out the bytes of code at the instruction pointer. Since the
    * crash may have been caused by an instruction pointer that was bad,
    * this code needs to be wrapped in an exception handler, in case there
    * is no memory to read. If the dereferencing of code[] fails, the
    * exception handler will print '??'.
    */
   hprintf(hLogFile, _T("\r\nBytes at CS:EIP:\r\n"));
   code = (BYTE *)Context->Eip;
   for (codebyte = 0; codebyte < NumCodeBytes; codebyte++)
   {
      __try
      {
         hprintf(hLogFile, _T("%02x "), code[codebyte]);

      }
      __except(EXCEPTION_EXECUTE_HANDLER)
      {
         hprintf(hLogFile, _T("?? "));
      }
   }

   /* Time to print part or all of the stack to the error log. This allows
    * us to figure out the call stack, parameters, local variables, etc.
    * Esp contains the bottom of the stack, or at least the bottom of
    * the currently used area
    */
   pStack = (DWORD *)Context->Esp;

   /* Unravel the stack callback chain */
   hprintf(hLogFile, _T("\n\nStack backtrace:\n"));

   pFrame = (DWORD *)Context->Ebp;
   while((pFrame) >= pStack && 
      (pFrame < (pStack + 0x00008000)))   /* Max of 32 K stack */
   {
      hprintf(hLogFile, _T(" %08x \n"), *(pFrame + 1));
      pFrame = (DWORD*)*pFrame;
   }


   DumpStack(hLogFile, pStack);

   DumpModuleList(hLogFile);

   hprintf(hLogFile, _T("\r\n===== [end of %s] =====\r\n"), XCRASHREPORT_ERROR_LOG_FILE);
   hflush(hLogFile);
   CloseHandle(hLogFile);

   /* Invoke pcaexcept.exe to send the exception dump log */
   WinExec("pcaexcept.exe", SW_SHOWNA);

   /* return the magic value which tells Win32 that this handler didn't
    * actually handle the exception - so that things will proceed as per
    * normal.
    */
   return EXCEPTION_EXECUTE_HANDLER ;
}


