/* winip.c
 * 
 *  Code to fish selected IP info from the Windows registry.
 *
 * Copyright 2003-2008 by Praemio Inc. All rights reserved.
 * Author: John Bartas - jbartas@praemio.com
 * This software is released under the terms of the LGPL.
 * See LGPL Notice at end of this file; Full license in ../license.txt
 */

#define _WINSOCK2API_   1

#include "busterport.h"
#include "brnet.h"
#include "ip.h"
#include "brdb.h"
#include "brdns.h"

#include "resource.h"

#define __MINGW32__ 1   /* required to omit IP v6 in pcap.h */
#include "pcap.h"
#include "brwinnet.h"


/* Varibles we set from the Windoze registery */
ip_addr  win_ip;        /* current IP address */
ip_addr  win_mask;      /* etc. */
ip_addr  win_router;
ip_addr  win_dns[DNSSERVERS];     /* these should be pretty static */


char * regIfaceBase;
char * regCtlBase;
char * regDhcpBase;     /* Windows 95-98 only */

extern   int   br_ifview;
extern   struct alarm_type    dead_iface;

extern  DWORD          dwVersion;     /* derivitives from osVer */
extern  DWORD          dwWindowsMajorVersion;


#ifndef MAX_PATH
#define MAX_PATH 256
#endif

char ifNameBuf[MAX_PATH];


/* br_GetRegIfIPaddr() - Helper routine to get registry IP info 
 * from a registry interface entry.
 *
 * Returns 0 if OK, else ENP_ error code.
 */

int
br_GetRegIfIPaddr(int iface,        /* IN - which interface */
                  char * regname,   /* IN - variable name */
                  HKEY ifKey,       /* IN - key for iface entry */
                  ip_addr * getip)
{
   char     varname[64];
   u_long   datatype;         /* for registry to return data type */
   u_long   buflen;           /* length of buffer, for pass to Reg routines */
   int      err;
   char *   space = NULL;

   if(strlen(regname) + 4 > sizeof(varname))
   {
      dtrap();
      return ENP_NOMEM;
   }
   /* get the active IP address from Windoze */
   if(nets[iface]->n_flags & NF_DHCPC)
      strcpy(varname, "Dhcp");
   else
      varname[0] = 0;
   
   strcat(varname, regname);

   buflen = MAX_PATH;
   err = RegQueryValueEx(ifKey, varname, NULL, 
         &datatype, (u_char*)ifNameBuf, &buflen);

   if(err)  /* Variable not found in registry for this interface */
   {
      /* This can happpen if the info is not yet set, for example 
       * we're probing an unplugged Ethernet interface for the 
       * gateway, and DHCP has (obviously) not set it. This is not
       * really an error, but not success either.
       */
      ifNameBuf[0] = 0;    /* clear buffer on error */
      return ENP_LOGIC;
   }

   /* if buffer has multiple addresses, get the first. Depending on the 
    * Windows version, they may be separated by spaces or commas. 
    */
   if((space = strchr(ifNameBuf, ' ')) != NULL)
      *space = 0;
   else if((space = strchr(ifNameBuf, ',')) != NULL)
      *space = 0;

   *getip = inet_addr(ifNameBuf);
   if(space)
      *space = ' ';
   return 0;
}

/* br_setipinfo(int iface)
 *
 * - Called to re-read and interface's IP info from system 
 *
 * The system should only update the info if it suspects the info
 * is currently inaccurate OR there is no old info.
 *
 * Returns 0 if OK, else ENP_ code
 */

int
br_setipinfo(int iface)
{
   HKEY     ifacesKey;        /* Key for "Interfaces" registry root */
   HKEY     ifKey;            /* reusable per-interface key */
   int      err = 0;          /* generic error */
   u_long   datatype;         /* for registry to return data type */
   u_long   buflen;           /* length of buffer, for pass to Reg routines */
   char *   cp;               /* scratch */
   char *   cp2;
   int      i;
   struct ip_hostentry * routerip;
   char *   clsId;            /* Registry "Class ID" name thingy */

   if((ndisifs[iface].flags & NF_WORKING) == 0)  /* if not up, return */
       return 0;

   /* get the correct registry strings for the Windoze flavor we're running on */
   if (!(dwVersion >= 0x80000000 && dwWindowsMajorVersion >= 4))
   { 
      /* Windows NT/2000/XP Version */
      regIfaceBase = "SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces";
   }
   else
   {
      dprintf("br_winp: no Win9X support\n");
      return ENP_LOGIC;
   }

   err = RegOpenKey(HKEY_LOCAL_MACHINE, regIfaceBase, &ifacesKey);
   if(err)
      return(ENP_LOGIC);

   clsId = &ndisifs[iface].name[12];
   err = RegOpenKey(ifacesKey, clsId, &ifKey);
   if(err)
   {
      RegCloseKey(ifacesKey);
      return(ENP_LOGIC);
   }
   buflen = MAX_PATH;
   err = RegQueryValueEx(ifKey, "EnableDHCP", NULL, 
            &datatype, (u_char*)ifNameBuf, &buflen);
   
   if(err || (*(DWORD*)ifNameBuf == 1))
      nets[iface]->n_flags |= NF_DHCPC;
   else
      nets[iface]->n_flags &= ~NF_DHCPC;


   /* Try for Windows IP addresses. We don't always get all of these, 
    *  depending on the system & network condiguration, so ignore
    * the error messages.
    */
   err = br_GetRegIfIPaddr(iface, "IPAddress", ifKey, &win_ip);
   err = br_GetRegIfIPaddr(iface, "SubnetMask", ifKey, &win_mask);
   err = br_GetRegIfIPaddr(iface, "DefaultGateway", ifKey, &win_router);
   err = br_GetRegIfIPaddr(iface, "NameServer", ifKey, &win_dns[0]);

   /* Get 2nd DNS server, if any */
   if(err == 0)
   {
      if(((cp = strchr(ifNameBuf, ',')) != NULL) ||
         ((cp = strchr(ifNameBuf, ' ')) != NULL))
      {
         cp2 = cp + 1;
         *cp = 0;
         win_dns[1] = inet_addr(cp2);
      }
   }
   RegCloseKey(ifKey);
   RegCloseKey(ifacesKey);

   /* At this point, the Winsock builds record the Winsock 
    * IP address as the Buster brstates[].br_ipaddr.
    */
   {
      /* If there is no IP address for the viewing iface, notify user. */
      if((win_ip == 0) && 
         (iface == br_ifview) &&
         (br_autoifview == FALSE))
      {
         ndisifs[iface].flags &= ~NF_WORKING;
         nets[iface]->ifOperStatus = 2;  /* DOWN*/

         alarm_general(&no_ipaddr, &brstates[iface].noip, 
            ndisifs[iface].description );
         return ENP_SENDERR;
      }

      /* If we have MAC information about the router then call
       * br_setrouter() with it, else just record the router 
       * IP address for now ands trust we'll find the MAC info 
       * later.
       */
      if(win_router)
      {
         routerip = br_iplookup(iface, win_router);
         if(routerip && routerip->mac_host)
            br_setrouter(iface, routerip->mac_host, 240);
         else
         {
            if(win_router != nets[iface]->n_router)
            {
               dprintf("br_setipinfo: Replacing nets[%d] router %u.%u.%u.%u ",
                  iface, PASS_IPADDR(nets[iface]->n_router) );
               dprintf("with router %u.%u.%u.%u from registry\n", 
                  PASS_IPADDR(win_router) );

               nets[iface]->n_router = win_router;
            }
         }
      }

      /* Put this info into our nets structure. If the the runtime 
       * structures are not set yet then put it into the static
       * startup copies.
       */
      if(nets[iface])
      {
         nets[iface]->n_ipaddr = win_ip;
         nets[iface]->n_mask = win_mask;
         nets[iface]->n_router = win_router;
      }
      else
      {
         staticnets[iface].n_ipaddr = win_ip;
         staticnets[iface].n_mask = win_mask;
         staticnets[iface].n_router = win_router;
      }

      /* Set flags if IP and router info are known */
      if(win_ip)
         brstates[iface].evflags |= EV_IPADDRSET;

      if(win_router)
         brstates[iface].evflags |= EV_ROUTERSET;

      if((win_ip == 0) || (win_router == 0))
      {
         nets[iface]->ifOperStatus = 2;  /* DOWN*/
         nets[iface]->ifLastChange = cticks * (100/TPS);
      }

      br_log(0, "Iface %d using Winsock IP address %u,%u,%u,%u\n", 
         iface, PASS_IPADDR(win_ip));
   }

   /* Set DNS servers */
   for(i = 0; i < DNSSERVERS; i++)
   {
      if(win_dns[i])
         br_adddnssvr(win_dns[i]);
      else
         break;
   }

   return 0;
}


void
nif_up(int iface)
{
   int   oldflags = ndisifs[iface].flags;

   ndisifs[iface].flags |= NF_WORKING;
   nets[iface]->ifOperStatus = 1;  /* UP */
   nets[iface]->ifLastChange = cticks * (100/TPS);

   /* If interface was down go (re)read IP info */
   if((oldflags & NF_WORKING) == 0)
      br_setipinfo(iface);

   return;
}

void
nif_down(int iface)
{
   int   i, working;

   ndisifs[iface].flags &= ~NF_WORKING;
   nets[iface]->ifOperStatus = 2;  /* DOWN*/
   nets[iface]->ifLastChange = cticks * (100/TPS);

   if(iface == br_ifview)
   {
      alarm_general(&dead_iface, &ndisifs[iface].ifalarm, 
         ndisifs[iface].description );
   }

   working = 0;
   for(i = 0; i < (int)ifNumber; i++)
   {
      if((nets[i]->ifOperStatus == 1) && 
         (nets[i]->n_router))
      {
         working++;
      }
   }
   if(working == 0)
      alarm_general(&no_iface, &no_if_inst, NULL );

   return;
}

/* br_nbiosocket()
 *
 * Portiability routine to set a socket in non-blocking mode.
 *
 * Windows (winsock) verison.
 */

int
br_nbiosocket(long socket)
{
   int   err;
   int   option = TRUE;

   err = ioctlsocket(socket, FIONBIO, (u_long *)&option);
   if(err)
   {
      /* get specific winsock error for debugger */
      err = WSAGetLastError();
      return ENP_LOGIC; 
   }
   return 0;
}

/*
Copyright (C) 2003-2008 by Praemio Inc. All rights reserved.

Author: John Bartas - jbartas@praemio.com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/


